def create_model(self, model_type=LogisticRegression, regression_targets=None, **kwargs):
    """
    Creates model explainer for a given model.

    Parameters
    --------
    model_type: sklearn class
        Model class in sklearn-api style (should have methods `fit`, `predict_proba`).
    regression_targets: dict, optional
        Mapping from ``index_col`` to regression target e.g. LTV of user. Default: ``None``
    kwargs:
        Parameters for model class that you use, also contains parameters for ``retention.extract_features()``.

    Returns
    --------
    Creates ModelDescriptor
    """
    if hasattr(self, 'datatype') and self.datatype == 'features':
        features = self._obj.copy()
    else:
        if 'ngram_range' not in kwargs:
            kwargs.update({'ngram_range': (1, 2)})
        features = self.extract_features(**kwargs)
    if regression_targets is not None:
        target = self.make_regression_targets(features, regression_targets)
    else:
        target = features.index.isin(self.get_positive_users(**kwargs))
    feature_range = kwargs.pop('ngram_range')
    mod = ModelDescriptor(model_type, features, target, feature_range=feature_range, **kwargs)
    return mod


@staticmethod
def make_regression_targets(features, regression_targets):
    """
    Creates target vector for given features.

    Parameters
    --------
    features: pd.DataFrame
        Feature matrix.
    regression_targets: dict
        Mapping from ``index_col`` to regression target, e.g. LTV of user.

    Returns
    --------
    List of targets alligned to feature matrix indices

    Return type
    -------

    """
    return [regression_targets.get(i) for i in features.index]


def _process_thr(self, data, thr, max_steps=30, mod=lambda x: x, **kwargs):
    f = data.index.str.startswith('Accumulated')
    if kwargs.get('targets', True):
        f |= data.index.isin(self.retention_config['target_event_list'])
    print("""
    Unused events on first {} steps:
        {}
    """.format(max_steps, '\n\t'.join(data[f].index.tolist())))
    return data.loc[(mod(data) >= thr).any(1) | f]


def _process_target_config(self, data, cfg, target):
    target = 'positive_target_event' if target.startswith('pos_') else 'negative_target_event'
    target = self.retention_config.get(target)
    for key, val in cfg.items():
        func = getattr(self, f'_process_{key}')
        data = func(data, val, target)
    return data


def _process_time_limit(self, data, threshold, name):
    self._init_cols(locals())
    if 'next_timestamp' in data:
        col = 'next_timestamp'
        change_next = True
        data[self._event_time_col()] \
            = pd.to_datetime(data[self._event_time_col()])
    else:
        col = self._event_time_col()
        change_next = False
    data[col] = pd.to_datetime(data[col])
    max_time = data[col].max()
    tmp = data.groupby(self._index_col()).tail(1)
    tmp = tmp[(max_time - tmp[col]).dt.total_seconds() > threshold]

    if change_next:
        tmp[self._event_col()] = tmp.next_event.values
        tmp.next_event = name
        tmp[self._event_time_col()] += timedelta(seconds=1)
        tmp['next_timestamp'] += timedelta(seconds=1)
    else:
        tmp[self._event_col()] = name
        tmp[self._event_time_col()] += timedelta(seconds=1)
    data.reset_index(drop=True, inplace=True)

    return data.append(tmp, ignore_index=True).reset_index(drop=True)


def _process_event_list(self, data, event_list, name):
    self._init_cols(locals())
    if 'next_event' in data:
        col = 'next_event'
    else:
        col = self._event_col()
    data[col] = np.where(data[col].isin(event_list), name, data[col])
    return data


def _process_empty(self, data, other, name):
    self._init_cols(locals())
    if 'next_event' in data:
        col = 'next_event'
        change_next = True
        data['next_timestamp'] \
            = pd.to_datetime(data[self._event_time_col()])
    else:
        col = self._event_col()
        change_next = False
    data[self._event_time_col()] \
        = pd.to_datetime(data[self._event_time_col()])
    bads = set(data[data[col] == other][self._index_col()])
    goods = set(data[self._index_col()]) - bads
    tmp = data[data[self._index_col()].isin(goods)]
    tmp = tmp.groupby(self._index_col()).tail(1)
    if change_next:
        tmp[self._event_col()] = tmp.next_event.values
        tmp.next_event = name
        tmp[self._event_time_col()] += timedelta(seconds=1)
        tmp['next_timestamp'] += timedelta(seconds=1)
    else:
        tmp[self._event_col()] = name
        tmp[self._event_time_col()] += timedelta(seconds=1)
    data.reset_index(drop=True, inplace=True)
    return data.append(tmp, ignore_index=True).reset_index(drop=True)


def _add_first_event(self, first_event):
    self._init_cols(locals())
    top1 = self._obj.groupby(self._index_col()).head(1)
    if 'next_event' in top1:
        top1.next_event = top1[self._event_col()].values
    top1[self._event_col()] = first_event
    top1[self._event_time_col()] -= timedelta(seconds=1)
    return top1.append(self._obj, ignore_index=True).reset_index(drop=True)


def _convert_timestamp(self, time_col=None):
    self._init_cols(locals())
    timestamp = self._obj[self._event_time_col()].iloc[0]
    if hasattr(timestamp, 'second'):
        return
    if type(timestamp) != str:
        l = len(str(int(timestamp)))
        self._obj[self._event_time_col()] *= 10 ** (19 - l)
    self._obj[self._event_time_col()] = pd.to_datetime(self._obj[self._event_time_col()])


def calculate_delays(self, plotting=True, time_col=None, index_col=None, event_col=None, bins=15, **kwargs):
    """
    Displays the logarithm of delay between ``time_col`` with the next value in nanoseconds as a histogram.

    Parameters
    --------
    plotting: bool, optional
        If ``True``, then histogram is plotted as a graph. Default: ``True``
    time_col: str, optional
        Name of custom time column for more information refer to ``init_config``. For instance, if in config you have defined ``event_time_col`` as ``server_timestamp``, but want to use function over ``user_timestamp``. By default the column defined in ``init_config`` will be used as ``time_col``.
    index_col: str, optional
        Name of custom index column, for more information refer to ``init_config``. For instance, if in config you have defined ``index_col`` as ``user_id``, but want to use function over sessions. By default the column defined in ``init_config`` will be used as ``index_col``.
    event_col: str, optional
        Name of custom event column, for more information refer to ``init_config``. For instance, you may want to aggregate some events or rename and use it as new event column. By default the column defined in ``init_config`` will be used as ``event_col``.
    bins: int, optional
        Number of bins for visualisation. Default: ``50``

    Returns
    -------
    Delays in seconds for each ``time_col``. Index is preserved as in original dataset.

    Return type
    -------
    List
    """
    self._init_cols(locals())
    data = self._get_shift(index_col=self._index_col(),
                           event_col=self._event_col()).copy()

    delays = np.log((data['next_timestamp'] - data[self._event_time_col()]) // pd.Timedelta('1s'))

    if plotting:
        fig, ax = plot.sns.mpl.pyplot.subplots(
            figsize=kwargs.get('figsize', (15, 7)))  # control figsize for proper display on large bin numbers
        _, bins, _ = plt.hist(delays[~np.isnan(delays) & ~np.isinf(delays)], bins=bins, log=True)
        if not kwargs.get('logvals', False):  # test & compare with logarithmic and normal
            plt.xticks(bins, np.around(np.exp(bins), 1))
        plt.show()

    return np.exp(delays)


def insert_sleep_events(self, events, delays=None, time_col=None, index_col=None, event_col=None):
    """
    Populates given dataset with sleep events representing time difference between occuring events. Note that this method is not inplace.

    Parameters
    --------
    events: dict
        Event name and log nanosecond ranges in the following structure: ``'event_name' : ['from_logtime', 'to_logtime']``. Keys of the dictionary are custom event names, while values are lists of two floats indicating start and end of time difference in lorarithm nanoseconds.
    delays: list
        Timestamp differences of each event with the next one. If ``None``, then uses ``BaseDataset.rete.calculate_delays()``. Default: ``None``
    time_col: str, optional
        Name of custom time column for more information refer to ``init_config``. For instance, if in config you have defined ``event_time_col`` as ``server_timestamp``, but want to use function over ``user_timestamp``. By default the column defined in ``init_config`` will be used as ``time_col``.
    index_col: str, optional
        Name of custom index column, for more information refer to ``init_config``. For instance, if in config you have defined ``index_col`` as ``user_id``, but want to use function over sessions. By default the column defined in ``init_config`` will be used as ``index_col``.
    event_col:  str, optional
        Name of custom event column, for more information refer to ``init_config``. For instance, you may want to aggregate some events or rename and use it as new event column. By default the column defined in ``init_config`` will be used as ``event_col``.

    Returns
    --------
    Original dataframe with inserted sleep events.

    Return type
    -------
    pd.DataFrame
    """

    self._init_cols(locals())

    if delays is None:
        delays = self.calculate_delays(False, self._event_time_col(), self._index_col(), self._event_col())

    data = self._obj.copy()
    to_add = []

    for event_name, (t_min, t_max) in events.items():
        tmp = data.loc[(delays >= t_min) & (delays < t_max)]
        tmp[self._event_col()] = event_name
        tmp[self._event_time_col()] += pd.Timedelta((np.e ** t_min) / 2)
        to_add.append(tmp)
        data['next_event'] = np.where((delays >= t_min) & (delays < t_max), event_name, data['next_event'])
        data['next_timestamp'] = np.where((delays >= t_min) & (delays < t_max),
                                          data[self._event_time_col()] + pd.Timedelta((np.e ** t_min) / 2),
                                          data['next_timestamp'])
    to_add.append(data)
    to_add = pd.concat(to_add)
    return to_add.sort_values(self._event_col()).reset_index(drop=True)


def remove_events(self, event_list, mode='equal'):
    """
    Removes events from dataset.

    Parameters
    --------
    event_list: list or str
        Events or other elements of ``event_col`` that should be filtered out.
    mode: str, optional
        Type of comparison:
            - `equal`: full event name match with element from ``event_list``;
            - `startswith`: event name starts with element from ``event_list``;
            - `contains`: event name contains element from ``event_list``.

    Returns
    --------
    Filtered dataframe based on event names.

    Return type
    -------
    pd.DataFrame
    """
    self._init_cols(locals())
    data = self._obj.copy()
    func = getattr(preprocessing, '_event_filter_' + mode)

    for event in event_list:
        data = data.loc[func(data[self._event_col()], event)]
    return data.reset_index(drop=True)


def select_bbox_from_tsne(self, bbox, plotting=True, **kwargs):
    """
    Selects data filtered by cordinates of TSNE plot.

    Parameters
    ---------
    bbox: list
        List of lists that contains angles of bbox.
            ```bbox = [
                [0, 0], # [min x, max x]
                [10, 10] # [min y, max y]
            ]```
    plotting: bool, optional
        If ``True``, then visualize graph of selected users.

    Returns
    --------
    Dataframe with filtered clickstream of users in bbox.

    Return type
    -------
    pd.DataFrame
    """
    self._init_cols(locals())
    if not hasattr(self, '_tsne'):
        raise ValueError('Please, use `learn_tsne` before selection of specific bbox')

    f = self._tsne.index.values[(self._tsne.iloc[:, 0] >= bbox[0][0])
                                & (self._tsne.iloc[:, 0] <= bbox[0][1])
                                & (self._tsne.iloc[:, 1] >= bbox[1][0])
                                & (self._tsne.iloc[:, 1] <= bbox[1][1])]

    filtered = self._obj[self._obj[self._index_col()].isin(f)]
    if plotting:
        filtered.rete.plot_graph(**kwargs)
    return filtered.reset_index(drop=True)


def show_tree_selector(self, **kwargs):
    """
    Shows tree selector for event filtering, based on values in ``event_col`` column. It uses `_` for event splitting and aggregation, so ideally the event name structure in the dataset should include underscores, e.g. ``[section]_[page]_[action]``. In this case event names are separated into levels, so that all the events with the same ``[section]`` will be placed under the same section, etc.
    There two kind of checkboxes in IFrame: large blue and small white. The former are used to include or exclude event from original dataset. The latter are used for event aggregation: toggle on a checkbox to aggregate all the underlying events to this level.
    Tree filter has a download button in the end of event list, which downloads a JSON config file, which you then need to use to filter and aggregate events with ``BaseDataset.rete.use_tree_filter()`` method.

    Parameters
    --------
    event_col: str, optional
        Name of custom event column, for more information refer to ``init_config``. For instance, you may want to aggregate some events or rename and use it as new event column. By default the column defined in ``init_config`` will be used as ``event_col``.
    width: int, optional
        Width of IFrame object with filters.
    height: int, optional
        Height of IFrame object with filters.

    Returns
    --------
    Renders events tree selector

    Return type
    --------
    IFrame
    """
    self._init_cols(locals())
    from retentioneering.core.tree_selector import show_tree_filter
    show_tree_filter(self._obj[self._event_col()], **kwargs)


def use_tree_filter(self, path, **kwargs):
    """
    Uses generated with ``show_tree_filter()`` JSON config to filter and aggregate ``event_col`` values of dataset.

    Parameters
    --------
    path: str
        Path to JSON config file generated with ``show_tree_filter()`` method.

    Returns
    --------
    Filtered and aggregated dataset

    Return type
    --------
    pd.DataFrame
    """
    from retentioneering.core.tree_selector import use_tree_filter
    res = use_tree_filter(self._obj, path, **kwargs)
    return res


def _create_bins(self, data, time_step, index_col=None):
    self._init_cols(locals())
    tmp = data.join(
        data.groupby(self._index_col())
        [self._event_time_col()].min(),
        on=self._index_col(), rsuffix='_min')

    data['bins'] = (
            data[self._event_time_col()] - tmp[self._event_time_col() + '_min']
    )
    data['bins'] = np.floor(data['bins'] / np.timedelta64(1, time_step))


def survival_curves(self, groups, spec_event=None, time_min=None, time_max=None, event_col=None, index_col=None,
                    target_event=None, time_step='D', plotting=True, **kwargs):
    """
    Plot survival curves for given grouping.

    Parameters
    --------
    groups: np.array
        Array of clickstream shape that splits data into different groups.
    spec_event: str, optional
        Event specific for test, e.g. we change auth flow, so we need to compare only users, who have started authorization, in this case `spec_event='auth_start'`. Default: ``None``
    time_min: int, optional
        Time when A/B test was started. If ``None``, then whole dataset is used. Defaul: ``None``
    time_max: int, optional
        Time when A/B test was ended. If ``None``, then whole dataset is used. Default: ``None``
    index_col: str, optional
        Name of custom index column, for more information refer to ``init_config``. For instance, if in config you have defined ``index_col`` as ``user_id``, but want to use function over sessions. By default the column defined in ``init_config`` will be used as ``index_col``.
    event_col: str, optional
        Name of custom event column, for more information refer to ``init_config``. For instance, you may want to aggregate some events or rename and use it as new event column. By default the column defined in ``init_config`` will be used as ``event_col``.
    target_event: str, optional
        Name of target event. If ``None``, then taken from ``retention_config``. Default: ``None``
    time_step: str, optional
        Time step for calculation of survival rate at specific time.
            Possible options:
                (`'D'` -- day, `'M'` -- month, `'h'` -- hour, `'m'` -- minute, `'Y'` -- year,
                 `'W'` -- week, `'s'` -- seconds, `'ms'` -- milliseconds).
        Default is day (`'D'`).
    plotting: bool, optional
        If ``True``, then plots survival curves.

    Returns
    --------
    Dataframe with points at survival curves and prints chi-squared LogRank test for equality statistics.

    Return type
    -------
    pd.DataFrame
    """
    self._init_cols(locals())
    data = self._obj.copy()
    if spec_event is not None:
        users = (data[data[self._event_col()] == spec_event][self._index_col()]).unique()
        f = data[self._index_col()].isin(users)
        data = data[f].copy()
        groups = groups[f].copy()
    if type(data[self._event_time_col()].iloc[0]) not in (int, float, object, str):
        data[self._event_time_col()] = pd.to_datetime(
            data[self._event_time_col()])
    if time_min is not None:
        f = data[self._event_time_col()] >= pd.to_datetime(time_min)
        data = data[f].copy()
        groups = groups[f].copy()
    if time_max is not None:
        f = data[self._event_time_col()] <= pd.to_datetime(time_max)
        data = data[f].copy()
        groups = groups[f].copy()
    self._create_bins(data, time_step, index_col)

    data['metric_col'] = (data
                          [self._event_col()]
                          == (target_event or self.retention_config['positive_target_event']))
    tmp = data[data.metric_col == 1]
    curves = tmp.groupby(
        [groups, 'bins']
    )[self._index_col()].nunique().rename('metric').reset_index()
    curves = curves.sort_values('bins', ascending=False)
    curves['metric'] = curves.groupby(groups.name).metric.cumsum()
    curves = curves.sort_values('bins')
    res = (curves
           .merge(curves
                  .groupby(groups.name)
                  .head(1)[[groups.name, 'metric']],
                  on=groups.name, suffixes=('', '_max')))
    self._logrank_test(res, groups.name)
    res['metric'] = res.metric / res.metric_max
    if plotting:
        plot.sns.lineplot(data=res, x='bins', y='metric', hue=groups.name)
    return res


@staticmethod
def _logrank_test(x, group_col):
    x['next_metric'] = x.groupby(group_col).metric.shift(-1)
    x['o'] = (x['metric'] - x['next_metric'])
    oj = x.groupby('bins').o.sum()
    nj = x.groupby('bins').metric.sum()
    exp = (oj / nj).rename('exp')
    x = x.join(exp, on='bins')
    x1 = x[x[group_col]]
    x1.index = x1.bins
    up = (x1.o - x1.exp * x1.metric).sum()
    var = ((oj * (x1.metric / nj) * (1 - x1.metric / nj) * (nj - oj)) / (nj - 1)).sum()
    z = up ** 2 / var

    from scipy.stats import chi2
    pval = 1 - chi2.cdf(z, df=1)
    print(f"""
    There is {'' if pval <= 0.05 else 'no '}significant difference
    log-rank chisq: {z}
    P-value: {pval}
    """)


def index_based_split(self, index_col=None, test_size=0.2, seed=0):
    """
    Splits dataset between train and test based on ``index_col``.

    Parameters
    -------
    index_col: str, optional
        Name of custom index column, for more information refer to ``init_config``. For instance, if in config you have defined ``index_col`` as ``user_id``, but want to use function over sessions. By default the column defined in ``init_config`` will be used as ``index_col``.
    test_size: float, optional
        Rate of test subsample from 0 to 1. Default: ``0.2``
    seed: int, optional
        Random seed number. Default: ``0``

    Returns
    -------
    Two dataframes: train and test.

    Return type
    -------
    pd.DataFrame
    """
    self._init_cols(locals())
    np.random.seed(seed)
    ids = np.random.permutation(self._obj[self._index_col()].unique())
    f = self._obj[self._index_col()].isin(ids[int(ids.shape[0] * test_size):])
    return self._obj[f].copy(), self._obj[~f].copy()


def step_matrix_bootstrap(self, n_samples=10, sample_size=None, sample_rate=1, random_state=0, **kwargs):
    """
    Estimates means and standard deviations of step matrix values with bootstrap.

    Parameters
    --------
    n_samples: int, optional
        Number of samples for bootstrap. Default: ``10``
    sample_size: int, optional
        Size of each subsample. Default: ``None``
    sample_rate: float, optional
        Rate of each subsample. Note that it cannot be used with ``sample_size``. Default: ``1``
    random_state: int, optional
        Random state for sampling. Default: ``0``
    kwargs: optional
        Arguments of ``BaseDataset.rete.get_step_matrix()``

    Returns
    --------
    Two dataframes: with mean and standard deviation values.

    Return type
    --------
    pd.DataFrame
    """
    self._init_cols(locals())
    res = []
    base = pd.DataFrame(0,
                        index=self._obj[self._event_col()].unique(),
                        columns=range(1, kwargs.get('max_steps') or 31)
                        )
    thr = kwargs.pop('thr', None)
    plot_type = kwargs.pop('plot_type', None)
    for i in range(n_samples):
        tmp = self._obj.sample(n=sample_size, frac=sample_rate, replace=True, random_state=random_state + i)
        tmp = (tmp.rete.get_step_matrix(plot_type=False, **kwargs) + base).fillna(0)
        tmp = tmp.loc[base.index.tolist()]
        res.append(tmp.values[:, :, np.newaxis])
    kwargs.update({'thr': thr})
    res = np.concatenate(res, axis=2)
    piv = pd.DataFrame(res.mean(2), index=base.index, columns=base.columns)
    stds = pd.DataFrame(res.std(2), index=base.index, columns=base.columns)

    if not kwargs.get('reverse'):
        for i in self.retention_config['target_event_list']:
            piv = piv.append(self._add_accums(piv, i))
    if kwargs.get('thr'):
        thr = kwargs.pop('thr')
        piv = self._process_thr(piv, thr, kwargs.get('max_steps' or 30), **kwargs)
    if kwargs.get('sorting'):
        piv = self._sort_matrix(piv)
    if not kwargs.get('for_diff'):
        if kwargs.get('reverse'):
            piv.columns = ['n'] + ['n - {}'.format(i - 1) for i in piv.columns[1:]]
    if plot_type:
        plot.step_matrix(
            piv.round(2),
            title=kwargs.get('title',
                             'Step matrix {}'
                             .format('reversed' if kwargs.get('reverse') else '')), **kwargs)
        plot.step_matrix(
            stds.round(3),
            title=kwargs.get('title',
                             'Step matrix std'), **kwargs)
    if kwargs.get('dt_means') is not None:
        means = np.array(self._obj.groupby('event_rank').apply(
            lambda x: (x.next_timestamp - x.event_timestamp).dt.total_seconds().mean()
        ))
        piv = pd.concat([piv, pd.DataFrame([means[:kwargs.get('max_steps' or 30)]],
                                           columns=piv.columns, index=['dt_mean'])])
    return piv, stds


def core_event_distribution(self, core_events, index_col=None, event_col=None,
                            thresh=None, plotting=True, use_greater=True, **kwargs):
    self._init_cols(locals())
    if type(core_events) == str:
        core_events = [core_events]
    self._obj['is_core_event'] = self._obj[self._event_col()].isin(core_events)
    rates = self._obj.groupby(self._index_col()).is_core_event.mean()
    if plotting:
        plot.core_event_dist(rates, thresh, **kwargs)
    if use_greater:
        f = set(rates[rates >= thresh].index.values)
    else:
        f = set(rates[rates < thresh].index.values)
    return self._obj[self._obj[self._index_col()].isin(f)].reset_index(drop=True)


def pairwise_time_distribution(self, event_order, time_col=None, index_col=None,
                               event_col=None, bins=100, limit=180, topk=3):
    self._init_cols(locals())
    if 'next_event' not in self._obj.columns:
        data = self._get_shift(index_col=index_col,
                               event_col=event_col).copy()

    data['time_diff'] = (data['next_timestamp'] - data[
        time_col or self.retention_config['event_time_col']]).dt.total_seconds()
    f_cur = data[self._event_col()] == event_order[0]
    f_next = data['next_event'] == event_order[1]
    s_next = data[f_cur & f_next].copy()
    s_cur = data[f_cur & (~f_next)].copy()

    s_cur.time_diff[s_cur.time_diff < limit].hist(alpha=0.5, log=True,
                                                  bins=bins, label='Others {:.2f}'.format(
            (s_cur.time_diff < limit).sum() / f_cur.sum()
        ))
    s_next.time_diff[s_next.time_diff < limit].hist(alpha=0.7, log=True,
                                                    bins=bins,
                                                    label='Selected event order {:.2f}'.format(
                                                        (s_next.time_diff < limit).sum() / f_cur.sum()
                                                    ))
    plot.sns.mpl.pyplot.legend()
    plot.sns.mpl.pyplot.show()
    (s_cur.next_event.value_counts() / f_cur.sum()).iloc[:topk].plot.bar()


@staticmethod
def _find_traj(x, event_list, event_col):
    res = np.ones_like(x[event_col]).astype(bool)
    for elem in event_list:
        res &= ((x[event_col] == elem).cumsum() > 0).values
    return res.max()


def create_trajectory_filter(self, event_list, index_col=None, event_col=None, **kwargs):
    self._init_cols(locals())
    df_stat = (self
               ._obj
               .groupby(self._index_col())
               .apply(self._find_traj,
                      event_list=event_list,
                      event_col=self._event_col()))
    return df_stat


def apply_trajectory_filter(self, event_list, index_col=None, event_col=None, **kwargs):
    self._init_cols(locals())
    f = self.create_trajectory_filter(event_list, index_col, event_col, **kwargs)
    f = self._obj[self._index_col()].isin(f[f].index.tolist())
    return self._obj[f].copy().reset_index(drop=True)


def _is_cycle(self, data):
    """
        Utilite for cycle search
    """
    temp = data.split('~~')
    return True if temp[0] == temp[-1] and len(set(temp)) > 1 else False


def _is_loop(self, data):
    """
        Utilite for loop search
    """
    temp = data.split('~~')
    return True if len(set(temp)) == 1 else False


def get_equal_fraction(self, fraction=1, random_state=42):
    """
        Selects fraction of good users and the same number of bad users

        Parameters
        --------
        fraction: float, optional
            Fraction of users. Should be in interval of (0,1]
        random_state: int, optional
            random state for numpy choice function

        Returns
        --------
        Two dataframes: with good and bad users

        Return type
        --------
        tuple of pd.DataFrame
    """
    if fraction <= 0 or fraction > 1:
        raise ValueError('The fraction is <= 0 or > 1')
    self._init_cols(locals())

    np.random.seed(random_state)
    good_users = self.get_positive_users()
    bad_users = self.get_negative_users()

    sample_size = min(int(len(good_users) * fraction), len(bad_users))
    good_users_sample = set(np.random.choice(good_users, sample_size, replace=False))
    bad_users_sample = set(np.random.choice(bad_users, sample_size, replace=False))

    return (self._obj[self._obj[self._index_col()].isin(good_users_sample)],
            self._obj[self._obj[self._index_col()].isin(bad_users_sample)])


def _remove_duplicates(self, data):
    """
    Removing same events, that are going one after another
    ('ev1 -> ev1 -> ev2 -> ev1 -> ev3 -> ev3   --------> ev1 -> ev2 -> ev1 -> ev3').
    This utilite is used in a find_sequences function

    """
    t = data.split('~~')
    t = '~~'.join([t[0]] + ['~~'.join(word for ind, word in enumerate(t[1:]) if t[ind] != t[ind + 1])])
    return t[:-2] if t[-1] == '~' else t


def find_sequences(self, ngram_range=(1, 1), fraction=1, random_state=42, exclude_cycles=False, exclude_loops=False,
                   exclude_repetitions=False, threshold=0, coefficient=0):
    """
        Finds all subsequences of length lying in interval

        Parameters
        --------
        fraction: float, optional
            Fraction of users. Should be in interval of (0,1]
        random_state: int, optional
            random state for numpy choice function

        Returns
        --------
        Two dataframes: with good and bad users

        Return type
        --------
        tuple of pd.DataFrame
    """
    self._init_cols(locals())
    sequences = dict()
    good, bad = self.get_equal_fraction(fraction, random_state)
    countvect = CountVectorizer(ngram_range=ngram_range, token_pattern='[^~]+')
    good_corpus = good.groupby(self._index_col())[self._event_col()].apply(
        lambda x: '~~'.join([l.lower() for l in x if l != 'pass' and l != 'lost']))
    good_count = countvect.fit_transform(good_corpus.values)
    good_frame = pd.DataFrame(columns=['~~'.join(x.split(' ')) for x in countvect.get_feature_names()],
                              data=good_count.todense())
    bad_corpus = bad.groupby(self._index_col())[self._event_col()].apply(
        lambda x: '~~'.join([l.lower() for l in x if l != 'pass' and l != 'lost']))
    bad_count = countvect.fit_transform(bad_corpus.values)
    bad_frame = pd.DataFrame(columns=['~~'.join(x.split(' ')) for x in countvect.get_feature_names()],
                             data=bad_count.todense())

    res = pd.concat([good_frame.sum(), bad_frame.sum()], axis=1).fillna(0).reset_index()
    res.columns = ['Sequence', 'Good', 'Lost']

    if exclude_cycles:
        res = res[~res.Sequence.apply(lambda x: self._is_cycle(x))]
    if exclude_loops:
        temp = res[~res.Sequence.apply(lambda x: self._is_loop(x))]
    if exclude_repetitions:
        res.Sequence = res.Sequence.apply(lambda x: self._remove_duplicates(x))
        res = res.groupby(res.Sequence)[['Good', 'Lost']].sum().reset_index()
        res = res[res.Sequence.apply(lambda x: len(x.split('~~')) in range(ngram_range[0], ngram_range[1] + 1))]

    res['Lost2Good'] = res['Lost'] / res['Good']
    return res[(abs(res['Lost2Good'] - 1) > coefficient) & (res.Good + res.Lost > threshold)] \
        .sort_values('Lost', ascending=False).reset_index(drop=True)


def find_cycles(self, interval, fraction=1, random_state=42, exclude_loops=False, exclude_repetitions=False):
    """

    Parameters
    ----------
    interval - interval of lengths for search. Any int number
    fraction - fraction of good users. Any float in (0,1]
    random_state - random_state for numpy random seed

    Returns pd.DataFrame with cycles
    -------

    """
    self._init_cols(locals())
    temp = self.find_sequences(interval, fraction, random_state, exclude_loops=exclude_loops,
                               exclude_repetitions=exclude_repetitions).reset_index(drop=True)
    return temp[temp['Sequence'].apply(lambda x: self._is_cycle(x))].reset_index(drop=True)


def find_loops(self, fraction=1, random_state=42):
    """
    Function for loop searching
    Parameters
    ----------
    fraction - fraction of good users. Any float in (0,1]
    random_state - random_state for numpy random seed

    Returns pd.DataFrame with loops. Good, Lost columns are for all occurences,
    (Good/Lost)_no_duplicates are for counting each cycle only once for user in which they occur
    -------

    """

    def loop_search(data, self_loops, event_list, is_bad):
        self._init_cols(locals())
        event_list = {k: 0 for k in event_list}
        for ind, url in enumerate(data[1:]):
            if data[ind] == data[ind + 1]:
                if url in self_loops.keys():
                    self_loops[url][is_bad] += 1
                    if event_list[url] == 0:
                        self_loops[url][is_bad + 3] += 1
                        event_list[url] = 1
                else:
                    self_loops[url] = [0, 0, 0, 0, 0, 0]
                    self_loops[url][is_bad] = 1
                    if event_list[url] == 0:
                        self_loops[url][is_bad + 3] += 1
                        event_list[url] = 1

    self._init_cols(locals())
    self_loops = dict()
    event_list = self._obj[self._event_col()].unique()
    good, bad = self.get_equal_fraction(fraction, random_state)
    for el in good.groupby(self._index_col()):
        loop_search(el[1][self._event_col()].values, self_loops, event_list, 0)

    for el in bad.groupby(self._index_col()):
        loop_search(el[1][self._event_col()].values, self_loops, event_list, 1)

    for key, val in self_loops.items():
        if val[0] != 0:
            self_loops[key][2] = val[1] / val[0]
        if val[3] != 0:
            self_loops[key][5] = val[4] / val[3]

    return pd.DataFrame(data=[[a[0]] + a[1] for a in self_loops.items()],
                        columns=['Sequence', 'Good', 'Lost', 'Lost2Good', 'GoodUnique',
                                 'LostUnique', 'UniqueLost2Good']) \
        .sort_values('Lost', ascending=False).reset_index(drop=True)



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


from datetime import datetime

import matplotlib.pyplot as plt
from matplotlib import rcParams
import pandas as pd
import seaborn as sns

from .plot_utils import __save_plot__, ___FigureWrapper__


@__save_plot__
def cluster_bar(data, *,
                clusters,
                target,
                target_names,
                plot_name=None,
                **kwargs):
    """
    Plots bar charts with cluster sizes and average target conversion rate.
    Parameters
    ----------
    data : pd.DataFrame
        Feature matrix.
    clusters : np.array
        Array of cluster IDs.
    target: np.array
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.

    target: list of np.arrays
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.


    plot_name : str, optional
        Name of plot to save. Default: ``'clusters_bar_{timestamp}.svg'``
    kwargs: optional
        Width and height of plot.
    Returns
    -------
    Saves plot to ``retention_config.experiments_folder``
    Return type
    -------
    PNG
    """
    cl = pd.DataFrame([clusters, *target], index=['clusters', *target_names]).T
    cl['cluster size'] = 1
    for t_n in target_names:
        cl[t_n] = cl[t_n].astype(int)

    bars = cl.groupby('clusters').agg({
        'cluster size': 'sum',
        **{t_n: 'mean' for t_n in target_names}
    }).reset_index()
    bars['cluster size'] /= bars['cluster size'].sum()

    bars = bars.melt('clusters', var_name='type', value_name='value')
    bars = bars[bars['type'] != ' '].copy()

    fig_x_size = round((1 + bars['clusters'].nunique()**0.7 * bars['type'].nunique()**0.7))
    rcParams['figure.figsize'] = fig_x_size, 6

    bar = sns.barplot(x='clusters', y='value', hue='type', data=bars)

    # move legend outside the box
    bar.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    y_value = ['{:,.2f}'.format(x * 100) + '%' for x in bar.get_yticks()]

    bar.set_yticks(bar.get_yticks().tolist())
    bar.set_yticklabels(y_value)
    bar.set(ylabel=None)

    # adjust the limits
    ymin, ymax = bar.get_ylim()
    if ymax > 1:
        bar.set_ylim(ymin, 1.05)

    plot_name = plot_name or 'cluster_bar_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.rete.retention_config['experiments_folder'] + '/' + plot_name
    return bar, plot_name, None, data.rete.retention_config


@__save_plot__
def cluster_event_dist(bars, *,
                       event_col,
                       cl1,
                       sizes,
                       cl2=None,
                       weight_col,
                       target_pos,
                       targets):
    fig_x_size = round(2 + (bars.shape[0] // 2) ** 0.8)
    rcParams['figure.figsize'] = fig_x_size, 6

    bar = sns.barplot(x=event_col, y='freq', hue='hue',
                      hue_order=[f'cluster {cl1}', 'all' if cl2 is None else f'cluster {cl2}'], data=bars)

    # move legend outside the box
    bar.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    y_value = ['{:,.2f}'.format(x * 100) + '%' for x in bar.get_yticks()]

    bar.set_yticks(bar.get_yticks().tolist())
    bar.set_yticklabels(y_value)
    bar.set_xticklabels(bar.get_xticklabels(), rotation=90)

    if weight_col is None:
        bar.set(ylabel='% from total events')
    else:
        bar.set(ylabel=f"% of '{weight_col}' with given event")
    bar.set(xlabel=None)

    # add vertical lines for central step-matrix
    if targets:
        bar.vlines([target_pos-0.52],
                  *bar.get_ylim(),
                  colors='Black',
                  linewidth=0.7,
                  linestyles='dashed')

    # adjust the limits
    ymin, ymax = bar.get_ylim()
    if ymax > 1:
        bar.set_ylim(ymin, 1.05)

    tit = f'top {bars.shape[0] // 2 - len(targets)} events in cluster {cl1} (size: {round(sizes[0] * 100, 2)}%) \n'
    tit += f'vs. all data (100%)' if cl2 is None else f'vs. cluster {cl2} (size: {round(sizes[1] * 100, 2)}%)'
    bar.set_title(tit)

    plot_name = 'cluster_event_dist_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = bars.rete.retention_config['experiments_folder'] + '/' + plot_name
    return bar, plot_name, None, bars.rete.retention_config


@__save_plot__
def core_event_dist(rates, thresh, plot_name=None, **kwargs):
    hist = sns.distplot(rates.values, hist=True, bins=kwargs.get('bins'), kde=kwargs.get('kde'))
    if thresh is not None:
        sns.mpl.pyplot.axvline(thresh, c='C1')

    plot_name = plot_name or 'core_event_dist_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    rates = rates.reset_index()
    plot_name = rates.rete.retention_config['experiments_folder'] + '/' + plot_name
    return hist, plot_name, None, rates.rete.retention_config


@__save_plot__
def permutation_importance(x, plot_name=None, **kwargs):
    fig = sns.mpl.pyplot.figure(figsize=[20, 7])
    sns.mpl.pyplot.xticks(rotation="vertical")
    sns.mpl.pyplot.title("Permutation importances")
    sns.mpl.pyplot.bar(x.feature.map(lambda x: " ".join(x)), x.importances_mean, yerr=x.importances_std)
    plot_name = plot_name or 'permutation_importance_{}'.format(datetime.now()).replace(':', '_').replace('.',
                                                                                                          '_') + '.svg'
    return ___FigureWrapper__(fig), plot_name, None, x.rete.retention_config










# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import pandas as pd
import numpy as np
from IPython.display import IFrame, display
import json

__TEMPLATE__ = """
<!DOCTYPE html>
<html>
<head>
  <title>Expand All. Collapse All</title>
  <meta charset="UTF-8">
  
  <!-- first import Vue -->
  <script src="https://unpkg.com/vue/dist/vue.js"></script>
  <!-- import JavaScript -->
  <script src="https://cdn.jsdelivr.net/npm/liquor-tree/dist/liquor-tree.umd.js"></script>
</head>
<body>
  <div id="app">
      <tree
        ref="tree"      
        :data="treeData"
        :options="treeOptions"
        @node:expanded="onNodeExpand">
        <div slot-scope="{{ node }}" class="node-container">
            <div class="node-controls">
              <input type="checkbox" @click.stop="aggregateTo(node)">{{{{node.text}}}}</input>
            </div>
        </div>
      </tree>
    <div id="download">
      <a href="#" @click.stop="saveFile()">Download Filter</a>
    </div>
  </div>

  <script>
    let aggList = [];

    new Vue({{
      el: '#app',
      data: function() {{
        return {{
          treeData: getTreeData(),
          treeOptions: {{
          	checkbox: true
          }}
        }}
      }},

      methods: {{
        aggregateTo: function(node) {{
            if (!aggList.includes(node.text)) {{
                aggList.push(node.text)
            }} else {{
                aggList = aggList.filter(function(value, index, arr) {{
                  return (value !== node.text)
                }})
            }}
            console.log(aggList)
        }},
      	onNodeExpand : function(node) {{
         const expandedNodes = this.$refs.tree.findAll({{ state: {{ expanded: true }} }});

         expandedNodes.forEach(expandedNode => {{
           if (expandedNode.id !== node.id && node.depth === expandedNode.depth) {{
             expandedNode.collapse()
           }}
         }});
        }},
        saveFile : function() {{
      	  let checkedNodes = this.$refs.tree.findAll({{ state: {{ checked: true }} }});
      	  let filter_names = [];
      	  checkedNodes.forEach(function (x) {{
      	    filter_names.push(x.data.text);
          }});
      	  downloadLayout({{
            "filter_names": filter_names,
            "agg_list": aggList
          }});
        }}
      }}
    }});


function getTreeData() {{
  return {tree_data}
}}

function downloadLayout(x) {{
    var a = document.createElement("a");
    var file = new Blob([JSON.stringify(x)], {{type: "text/json;charset=utf-8"}});
    a.href = URL.createObjectURL(file);
    a.download = "filter_list.json";
    a.click();
}}
  </script>
</body>
</html>
"""


def _create_node(idx, splitted_names, res, pre):
    if idx == splitted_names.shape[1]:
        return res
    cols = splitted_names[idx].unique()
    for i in cols:
        if i is None:
            continue
        res.append({
            'text': pre + i,
            'state': {
                'expanded': False,
                'checked': True,
            },
            'children': _create_node(idx + 1, splitted_names[splitted_names[idx] == i], [], pre + i + '_')
        })
    return res


def show_tree_filter(event_col, width=500, height=500, **kwargs):
    """
    Shows tree selector for events filtering and aggregation, based on values in ``event_col`` column. It uses `_` for event names splitting, so ideally the event name structure in the dataset should include underscores and have a hierarchical, e.g. ``[section]_[page]_[action]``. In this case event names are separated into levels, so that all the events with the same ``[section]`` will be placed under the same tree node, etc.
    There two kind of checkboxes in tree selector: large blue and small white. The former are used to include or exclude event from original dataset. The latter are used for event aggregation: toggle on a checkbox against an event name level, e.g. a specific ``[section]_[page]``, to aggregate all the underlying events, all the ``[actions]`` in this example, to this level.
    Tree filter has a download button in the end of event list, which downloads a JSON config file, which you then need to use to filter and aggregate events with ``use_tree_filter()`` function.

    Parameters
    --------
    event_col: str
        Column with event names.
    width: int, optional
        Width of IFrame in pixels.
    height: int, optional
        Height of IFrame in pixels.

    Returns
    --------
    Renders events tree selector

    Return type
    --------
    IFrame
    """

    splitted_names = pd.Series(event_col.unique()).str.split('_', expand=True)
    res = _create_node(0, splitted_names, [], '')
    res = __TEMPLATE__.format(tree_data=json.dumps(res))
    with open('./filter.html', 'w') as f:
        f.write(res)
    display(IFrame('./filter.html', width=width, height=height))


def use_tree_filter(data, path, **kwargs):
    """
    Uses generated with ``show_tree_filter()`` JSON config to filter and aggregate ``event_col`` values of dataset.

    Parameters
    --------
    data: pd.DataFrame
        Event dataset. Should include the same event column as in ``event_col`` of ``retention_config``.
    path: str
        Path to JSON config file generated with ``show_tree_filter()`` function.

    Returns
    --------
    Filtered and aggregated dataset

    Return type
    --------
    pd.DataFrame
    """
    with open(path) as f:
        tree_filter = json.load(f)

    f = data[data.rete.retention_config['event_col']].isin(tree_filter['filter_names'])
    data = data[f].reset_index(drop=True)
    data['dump_event_col'] = data[data.rete.retention_config['event_col']]
    for agg in tree_filter['agg_list']:
        f = data[data.rete.retention_config['event_col']].str.startswith(agg)
        data[data.rete.retention_config['event_col']] = np.where(f, agg,
                                                                      data[data.rete.retention_config['event_col']]
                                                                      )
    return data



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import retentioneering
from retentioneering import datasets
from retentioneering.datasets import gen_corner_cases
import operator

# ******************************
# **** INITIALIZE dataset 1 ****
# ******************************

simple_shop_data = datasets.load_simple_shop()


retentioneering.config.update({
    'event_col': 'event',
    'event_time_col': 'timestamp',
    'user_col': 'user_id'
})


# **********************************
# **** ADD corner cases to list ****
# **********************************

# list of parameters to test:
# add non-corner case:
test_datasets = [dict(test_dataset=simple_shop_data)]


# add corner cases to list of cases
for case in gen_corner_cases.__all__:
    # extract function from module by name
    func = operator.attrgetter(case)(gen_corner_cases)
    test_datasets.append(dict(test_dataset=func(simple_shop_data)))



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import pandas as pd


def get_edgelist(self, *,
                 weight_col=None,
                 norm_type=None,
                 edge_attributes='edge_weight'):
    """
    Creates weighted table of the transitions between events.

    Parameters
    ----------
    weight_col: str (optional, default=None)
        Aggregation column for transitions weighting. To calculate weights
        as number of transion events use None. To calculate number
        of unique users passed through given transition 'user_id'.
         For any other aggreagtion, like number of sessions, pass the column name.

    norm_type: {None, 'full', 'node'} (optional, default=None)
        Type of normalization. If None return raw number of transtions
        or other selected aggregation column. 'full' - normalized over
        entire dataset. 'node' weight for edge A --> B normalized over
        user in A

    edge_attributes: str (optional, default 'edge_weight')
        Name for edge_weight columns

    Returns
    -------
    Dataframe with number of rows equal to all transitions with weight
    non-zero weight

    Return type
    -----------
    pd.DataFrame
    """
    if norm_type not in [None, 'full', 'node']:
        raise ValueError(f'unknown normalization type: {norm_type}')

    event_col = self.retention_config['event_col']
    time_col = self.retention_config['event_time_col']

    cols = [event_col, 'next_' + str(event_col)]

    data = self._get_shift().copy()

    # get aggregation:
    if weight_col is None:
        agg = (data
               .groupby(cols)[time_col]
               .count()
               .reset_index())
        agg.rename(columns={time_col: edge_attributes}, inplace=True)
    else:
        agg = (data
               .groupby(cols)[weight_col]
               .nunique()
               .reset_index())
        agg.rename(columns={weight_col: edge_attributes}, inplace=True)

    # apply normalization:
    if norm_type == 'full':
        if weight_col is None:
            agg[edge_attributes] /= agg[edge_attributes].sum()
        else:
            agg[edge_attributes] /= data[weight_col].nunique()

    if norm_type == 'node':
        if weight_col is None:
            event_transitions_counter = data.groupby(event_col)[cols[1]].count().to_dict()
            agg[edge_attributes] /= agg[cols[0]].map(event_transitions_counter)
        else:
            user_counter = data.groupby(cols[0])[weight_col].nunique().to_dict()
            agg[edge_attributes] /= agg[cols[0]].map(user_counter)

    return agg



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


__all__ = ['keep_one_user',
           'keep_one_event']


def keep_one_user(dataset):
    index_col = dataset.rete.retention_config['user_col']

    df = dataset.copy()

    user_to_keep = df[index_col][0]
    df = df[df[index_col] == user_to_keep].copy()

    return df


def keep_one_event(dataset):
    event_col = dataset.rete.retention_config['event_col']

    df = dataset.copy()

    event_to_keep = df[event_col][0]
    df = df[df[event_col] == event_to_keep].copy()

    return df



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import pandas as pd

from retentioneering.visualization import plot_project
from .extract_features import _learn_tsne, _learn_umap


def project(self, *,
            method='tsne',
            targets=(),
            ngram_range=(1,1),
            feature_type='tfidf',
            plot_type=None,
            **kwargs):
    """
    Does dimention reduction of user trajectories and draws projection plane.

    Parameters
    ----------
    method: {'umap', 'tsne'} (optional, default 'tsne')
        Type of manifold transformation.
    plot_type: {'targets', 'clusters', None} (optional, default None)
        Type of color-coding used for projection visualization:
            - 'clusters': colors trajectories with different colors depending on cluster number.
            IMPORTANT: must do .rete.get_clusters() before to obtain cluster mapping.
            - 'targets': color trajectories based on reach to any event provided in 'targets' parameter.
            Must provide 'targets' parameter in this case.
        If None, then only calculates TSNE without visualization.

    targets: list or tuple of str (optional, default  ())
        Vector of event_names as str. If user reach any of the specified events, the dot corresponding
        to this user will be highlighted as converted on the resulting projection plot

    feature_type: str, (optional, default 'tfidf')
        Type of vectorizer to use before dimension-reduction. Available vectorization methods:
        {'tfidf', 'count', 'binary', 'frequency'}

    ngram_range: tuple, (optional, default (1,1))
        The lower and upper boundary of the range of n-values for different
        word n-grams or char n-grams to be extracted before dimension-reduction.
        For example ngram_range=(1, 1) means only single events, (1, 2) means single events
        and bigrams.

    Returns
    --------
    Dataframe with data in the low-dimensional space for user trajectories indexed by user IDs.

    Return type
    --------
    pd.DataFrame
    """
    event_col = self.retention_config['event_col']
    index_col = self.retention_config['user_col']

    if plot_type == 'clusters':
        if hasattr(self, 'clusters'):
            targets_mapping = self.clusters
            legend_title = 'cluster number:'
        else:
            raise AttributeError("Run .rete.get_clusters() before using plot_type='clusters' to obtain clusters mapping")

    elif plot_type == 'targets':
        if targets is None:
            raise ValueError("When plot_type ='targets' must provide parameter targets as list of target event names")
        else:
            targets = [list(pd.core.common.flatten(targets))]
            legend_title = 'conversion to (' + ' | '.join(targets[0]).strip(' | ') + '):'
            targets_mapping = (self._obj
                                     .groupby(index_col)[event_col]
                                     .apply(lambda x: bool(set(*targets) & set(x)))
                                     .to_frame()
                                     .sort_index()[event_col]
                                     .values)

    features = self.extract_features(feature_type=feature_type,
                                     ngram_range=ngram_range)

    if method == 'tsne':
        self._projection = _learn_tsne(features, **kwargs)
    elif method == 'umap':
        self._projection = _learn_umap(features, **kwargs)

    # return only embeddings is no plot_type:
    if plot_type is None:
        return self._projection

    plot_project.plot_projection(
        projection=self._projection.values,
        targets=targets_mapping,
        legend_title=legend_title,
    )

    return self._projection



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from ...visualization.plot_funnel import plot_stacked_funnel


def funnel(self, *,
           targets,
           groups=None,
           group_names=None):
    """
    Plots simple convertion funnel with stages as specified in targets parameter.

    Parameters
    ----------
    targets: list of str
        List of events used as stages for the funnel. Absolute and relative
        number of users who reached specified events at least once will be
        plotted. Multiple events can be grouped together as individual state
        by combining them as sub list.
    groups: list of collectibles (optional, default None)
        List of user_ids collections. Funnel for each user_id collection
        will be plotted. If None all users from dataset will be plotted
    group_names: list of strings (optional, default None)
        Names for specified user groups to place in a legend. If specified
         len(group_names) must be equal to len(groups).

    Returns
    -------
    Funnel plot

    """

    data = self._obj
    event_col = self.retention_config['event_col']
    index_col = self.retention_config['user_col']

    # if group not specified select all users
    if groups is None:
        groups = [data[index_col].unique()]
        group_names = ['all users']
    elif group_names is None:
        group_names = [f"group {i}" for i in range(len(groups))]

    # pre-process targets

    # format targets to list of lists:
    for n, i in enumerate(targets):
        if type(i) != list:
            targets[n] = [i]

    # generate target_names:
    target_names = []
    for t in targets:
        # get name
        target_names.append(' | '.join(t).strip(' | '))

    res_dict = {}
    for group, group_name in zip(groups, group_names):

        # isolate users from group
        group_data = data[data[index_col].isin(group)]

        vals = []
        for target in targets:
            # define how many users have particular target:
            vals.append(group_data[group_data[event_col].isin(target)][index_col].nunique())
        res_dict[group_name] = {'targets': target_names, 'values': vals}

    return plot_stacked_funnel(res_dict)


# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import pandas as pd

from retentioneering.visualization import plot_step_matrix


def step_matrix(self, *,
                max_steps=20,
                weight_col=None,
                precision=2,
                targets=None,
                accumulated=None,
                sorting=None,
                thresh=0,
                centered=None,
                groups=None,
                show_plot=True):
    """
    Plots heatmap with distribution of users over trajectory steps ordered by
    event name. Matrix rows are event names, columns are aligned user trajectory
    step numbers and the values are shares of users. A given entry X at column i
    and event j means at i'th step fraction of users X  have specific event j.

    Parameters
    ----------
    max_steps: int (optional, default 20)
        Maximum number of steps in trajectory to include.

    weight_col: str (optional, default None)
        Aggregation column for edge weighting. If None, specified index_col
        from retentioneering.config will be used as column name. For example,
        can be specified as `session_id` if dataframe has such column.

    precision: int (optional, default 2)
        Number of decimal digits after 0 to show as fractions in the heatmap.

    thresh: float (optional, default 0)
        Used to remove rare events. Aggregates all rows where all values are
        less then specified threshold.

    targets: list (optional, default None)
        List of events names (as str) to include in the bottom of
        step_matrix as individual rows. Each specified target will have
        separate color-coding space for clear visualization. Example:
        ['product_page', 'cart', 'payment']. If multiple targets need to
        be compared and plotted using same color-coding scale, such targets
        must be combined in sub-list.
        Examples: ['product_page', ['cart', 'payment']]

    accumulated: string (optional, default None)
        Option to include accumulated values for targets. Valid values are
        None (do not show accumulated tartes), 'both' (show step values and
        accumulated values), 'only' (show targets only as accumulated).

    centered: dict (optional, default None)
        Parameter used to align user trajectories at specific event at specific
        step. Has to contain three keys:
            'event': str, name of event to align
            'left_gap': int, number of events to include before specified event
            'occurrence': int which occurance of event to align (typical 1)
        When this parameter is not None only users which have specified i'th
        'occurance' of selected event preset in their trajectories will
        be included. Fraction of such remaining users is specified in the title of
        centered step_matrix. Example:
        {'event': 'cart', 'left_gap': 8, 'occurrence': 1}

    sorting: list (optional, default None)
        List of events_names (as string) can be passed to plot step_matrix with
        specified ordering of events. If None rows will be ordered according
        to i`th value (first row, where 1st element is max, second row, where
        second element is max, etc)

    groups: tuple (optional, default None)
        Can be specified to plot step differential step_matrix. Must contain
        tuple of two elements (g_1, g_2): where g_1 and g_2 are collections
        of user_id`s (list, tuple or set). Two separate step_matrixes M1 and M2
        will be calculated for users from g_1 and g_2, respectively. Resulting
        matrix will be the matrix M = M1-M2. Note, that values in each column
        in differential step matrix will sum up to 0 (since columns in both M1
        and M2 always sum up to 1).

    show_plot: bool (optional, default True)
        whether to show resulting heatmap or not.

    Returns
    -------
    Dataframe with max_steps number of columns and len(event_col.unique)
    number of rows at max, or less if used thr > 0.

    Return type
    -----------
    pd.DataFrame
    """

    event_col = self.retention_config['event_col']
    weight_col = weight_col or self.retention_config['user_col']
    time_col = self.retention_config['event_time_col']

    data = self._obj.copy()

    # add termination event to each history:
    data = data.rete.split_sessions(thresh=None,
                                    eos_event='ENDED',
                                    session_col=None)

    data['event_rank'] = 1
    data['event_rank'] = data.groupby(weight_col)['event_rank'].cumsum()

    # BY HERE WE NEED TO OBTAIN FINAL DIFF piv and piv_targets before sorting, thresholding and plotting:

    if groups:
        data_pos = data[data[weight_col].isin(groups[0])].copy()
        if len(data_pos) == 0:
            raise IndexError('Users from positive group are not present in dataset')
        piv_pos, piv_targets_pos, fraction_title, window, targets_plot = \
            _step_matrix_values(data=data_pos,
                                weight_col=weight_col,
                                event_col=event_col,
                                time_col=time_col,
                                targets=targets,
                                accumulated=accumulated,
                                centered=centered,
                                max_steps=max_steps)

        data_neg = data[data[weight_col].isin(groups[1])].copy()
        if len(data_pos) == 0:
            raise IndexError('Users from negative group are not present in dataset')
        piv_neg, piv_targets_neg, fraction_title, window, targets_plot = \
            _step_matrix_values(data=data_neg,
                                weight_col=weight_col,
                                event_col=event_col,
                                time_col=time_col,
                                targets=targets,
                                accumulated=accumulated,
                                centered=centered,
                                max_steps=max_steps)

        def join_index(df1, df2):
            full_list = set(df1.index) | set(df2.index)
            for i in full_list:
                if i not in df1.index:
                    df1.loc[i] = 0
                if i not in df2.index:
                    df2.loc[i] = 0

        join_index(piv_pos, piv_neg)
        piv = piv_pos - piv_neg

        if targets:
            join_index(piv_targets_pos, piv_targets_neg)
            piv_targets = piv_targets_pos - piv_targets_neg
        else:
            piv_targets = None

    else:
        piv, piv_targets, fraction_title, window, targets_plot = \
            _step_matrix_values(data=data,
                                weight_col=weight_col,
                                event_col=event_col,
                                time_col=time_col,
                                targets=targets,
                                accumulated=accumulated,
                                centered=centered,
                                max_steps=max_steps)

    thresh_index = 'THRESHOLDED_'
    if thresh != 0:
        # find if there are any rows to threshold:
        thresholded = piv.loc[(piv.abs() < thresh).all(1)].copy()
        if len(thresholded) > 0:
            piv = piv.loc[(piv.abs() >= thresh).any(1)].copy()
            thresh_index = f'THRESHOLDED_{len(thresholded)}'
            piv.loc[thresh_index] = thresholded.sum()

    if sorting is None:
        piv = _sort_matrix(piv)

        keep_in_the_end = []
        keep_in_the_end.append('ENDED') if ('ENDED' in piv.index) else None
        keep_in_the_end.append(thresh_index) if (thresh_index in piv.index) else None

        events_order = [*(i for i in piv.index if i not in keep_in_the_end), *keep_in_the_end]
        piv = piv.loc[events_order]

        # add sorting list to config
        self.retention_config['step_matrix'] = {'sorting': events_order}

    else:
        # if custom sorting was provided:
        if not isinstance(sorting, list):
            raise TypeError('parameter `sorting` must be a list of event names')
        if {*sorting} != {*piv.index}:
            raise ValueError(
                'provided sorting list does not match list of events. Run with `sorting` = None to get the actual list')

        piv = piv.loc[sorting]

    if centered:
        piv.columns = [f'{int(i) - window - 1}' for i in piv.columns]
        if targets:
            piv_targets.columns = [f'{int(i) - window - 1}' for i in piv_targets.columns]

    if show_plot:
        plot_step_matrix.step_matrix(piv, piv_targets,
                                     targets_list=targets_plot,
                                     title=f'{"centered" if centered else ""} {"differential " if groups else ""}step matrix {fraction_title}',
                                     centered_position=window,
                                     precision=precision)

    return piv


def _step_matrix_values(*, data, weight_col, event_col, time_col,
                       targets, accumulated, centered, max_steps):
    """
    Supplemental function to calculate values for step_matrix

    Parameters same as in step_matrix

    Parameters
    ----------
    data
    weight_col
    event_col
    time_col
    targets
    accumulated
    centered
    max_steps

    Returns
    -------

    pandas Dataframe

    """

    def pad_cols(df, max_steps):
        """
        Parameters
        ----------
        df - dataframe
        max_steps - number of cols
        Returns
        -------
        returns Dataframe with columns from 0 to max_steps
        """
        df = df.copy()
        if max(df.columns) < max_steps:
            for col in range(max(df.columns) + 1, max_steps + 1):
                df[col] = 0
        # add missing cols if needed:
        if min(df.columns) > 1:
            for col in range(1, min(df.columns)):
                df[col] = 0
        # sort cols
        return df[list(range(1, max_steps + 1))]

    from copy import deepcopy

    data = data.copy()
    targets = deepcopy(targets)

    # ALIGN DATA IF CENTRAL
    fraction_title = ''
    window = None
    if centered is not None:
        # CHECKS
        if (not isinstance(centered, dict) or
                ({'event', 'left_gap', 'occurrence'} - {*centered.keys()})):
            raise ValueError("Parameter centered must be dict with following keys: 'event', 'left_gap', 'occurrence'")

        center_event = centered.get('event')
        window = centered.get('left_gap')
        occurrence = centered.get('occurrence')
        if occurrence < 1 or not isinstance(occurrence, int):
            raise ValueError("key value 'occurrence' must be Int >=1")
        if window < 0 or not isinstance(window, int):
            raise ValueError("key value 'left_gap' must be Int >=0")
        if center_event not in data[event_col].unique():
            raise ValueError(f'Event "{center_event}" not found in the column: "{event_col}"')

        # keep only users who have center_event at least N = occurrence times
        data['occurrence'] = data[event_col] == center_event
        data['occurance_counter'] = data.groupby(weight_col)['occurrence'].cumsum() * data['occurrence']
        users_to_keep = data[data['occurance_counter'] == occurrence][weight_col].unique()

        if len(users_to_keep) == 0:
            raise ValueError(f'no records found with event "{center_event}" occuring N={occurrence} times')

        fraction_used = len(users_to_keep) / data[weight_col].nunique() * 100
        if fraction_used < 100:
            fraction_title = f'({fraction_used:.1f}% of total records)'
        data = data[data[weight_col].isin(users_to_keep)].copy()

        def pad_to_center(x):
            position = x.loc[(x[event_col] == center_event) &
                             (x['occurance_counter'] == occurrence)]['event_rank'].min()
            shift = position - window - 1
            x['event_rank'] = x['event_rank'] - shift
            return x
        data = data.groupby(weight_col).apply(pad_to_center)
        data = data[data['event_rank'] > 0].copy()

    # calculate step matrix elements:
    agg = (data
           .groupby(['event_rank', event_col])[weight_col]
           .nunique()
           .reset_index())
    agg[weight_col] /= data[weight_col].nunique()

    agg = agg[agg['event_rank'] <= max_steps]
    agg.columns = ['event_rank', 'event_name', 'freq']

    piv = agg.pivot(index='event_name', columns='event_rank', values='freq').fillna(0)

    # add missing cols if number of events < max_steps:
    piv = pad_cols(piv, max_steps)

    piv.columns.name = None
    piv.index.name = None

    # MAKE TERMINATED STATE ACCUMULATED:
    if 'ENDED' in piv.index:
        piv.loc['ENDED'] = piv.loc['ENDED'].cumsum().fillna(0)

    # add NOT_STARTED events for centered matrix
    if centered:
        piv.loc['NOT_STARTED'] = 1 - piv.sum()

    # ADD ROWS FOR TARGETS:
    piv_targets = None
    if targets:
        # obtain flatten list of targets:
        targets_flatten = list(pd.core.common.flatten(targets))

        # format targets to list of lists:
        for n, i in enumerate(targets):
            if type(i) != list:
                targets[n] = [i]

        agg_targets = (data
                       .groupby(['event_rank', event_col])[time_col]
                       .count()
                       .reset_index())
        agg_targets[time_col] /= data[weight_col].nunique()
        agg_targets.columns = ['event_rank', 'event_name', 'freq']

        agg_targets = agg_targets[agg_targets['event_rank'] <= max_steps]

        piv_targets = agg_targets.pivot(index='event_name', columns='event_rank', values='freq').fillna(0)
        piv_targets = pad_cols(piv_targets, max_steps)

        # if target is not present in dataset add zeros:
        for i in targets_flatten:
            if i not in piv_targets.index:
                piv_targets.loc[i] = 0

        piv_targets = piv_targets.loc[targets_flatten].copy()

        piv_targets.columns.name = None
        piv_targets.index.name = None

        if accumulated == 'only':
            piv_targets.index = map(lambda x: 'ACC_' + x, piv_targets.index)
            for i in piv_targets.index:
                piv_targets.loc[i] = piv_targets.loc[i].cumsum().fillna(0)

            # change names is targets list:
            for target in targets:
                for j, item in enumerate(target):
                    target[j] = 'ACC_'+item

        if accumulated == 'both':
            for i in piv_targets.index:
                piv_targets.loc['ACC_'+i] = piv_targets.loc[i].cumsum().fillna(0)

            # add accumulated targets to the list:
            targets_not_acc = deepcopy(targets)
            for target in targets:
                for j, item in enumerate(target):
                    target[j] = 'ACC_'+item
            targets = targets_not_acc + targets

    return piv, piv_targets, fraction_title, window, targets


def _sort_matrix(step_matrix):
    x = step_matrix.copy()
    order = []
    for i in x.columns:
        new_r = x[i].idxmax()
        order.append(new_r)
        x = x.drop(new_r)
        if x.shape[0] == 0:
            break
    order.extend(list(set(step_matrix.index) - set(order)))
    return step_matrix.loc[order]



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from datetime import datetime

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import rcParams
import matplotlib.pylab as plt

from .plot_utils import __save_plot__


_ = pd.DataFrame()


@__save_plot__
def plot_projection(*,
                    projection,
                    targets,
                    legend_title):
    rcParams['figure.figsize'] = 8, 6

    scatter = sns.scatterplot(x=projection[:, 0],
                              y=projection[:, 1],
                              hue=targets,
                              legend='full',
                              palette=sns.color_palette("bright")[0:np.unique(targets).shape[0]])

    # move legend outside the box
    scatter.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.).set_title(legend_title)
    plt.setp(scatter.get_legend().get_title(), fontsize='12')

    plot_name = 'project_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = _.rete.retention_config['experiments_folder'] + '/' + plot_name

    return scatter, plot_name, projection, _.rete.retention_config


# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md



from math import isclose
from .prepare_test_datasets import test_datasets

# global constants
REL_TOL = 0.0001
ABS_TOL = 0.0001


# util function to cycle through parameters for each test
def pytest_generate_tests(metafunc):
    # called once per each test function
    funcarglist = metafunc.cls.params[metafunc.function.__name__]
    argnames = sorted(funcarglist[0])
    metafunc.parametrize(
        argnames, [[funcargs[name] for name in argnames] for funcargs in funcarglist]
    )


event_col = test_datasets[0]['test_dataset'].rete.retention_config['event_col']
next_event_col = 'next_' + event_col
time_col = test_datasets[0]['test_dataset'].rete.retention_config['event_time_col']
index_col = test_datasets[0]['test_dataset'].rete.retention_config['user_col']


# **********************
# **** define tests ****
# **********************

class TestNormalizationFuncs:
    # for each test we would like to run all cases
    params = {}.fromkeys(["test_no_norm_by_events",
                          "test_full_norm_by_events",
                          "test_node_norm_by_events",
                          "test_no_norm_by_users",
                          "test_full_norm_by_users",
                          "test_node_norm_by_users"], test_datasets)

    def test_no_norm_by_events(self, test_dataset):
        assert no_norm_by_events(test_dataset)

    def test_full_norm_by_events(self, test_dataset):
        assert full_norm_by_events(test_dataset)

    def test_node_norm_by_events(self, test_dataset):
        assert node_norm_by_events(test_dataset)

    def test_no_norm_by_users(self, test_dataset):
        assert no_norm_by_users(test_dataset)

    def test_full_norm_by_users(self, test_dataset):
        assert full_norm_by_users(test_dataset)

    def test_node_norm_by_users(self, test_dataset):
        assert node_norm_by_users(test_dataset)


# *****************************
# *** FUNCTIONS DEFINITIONS ***
# *****************************


def dict_compare(d1, d2):
    """
    Compare equality of two dictionaries with numeric values

    Parameters
    ----------
    d1, d2 - dictionaries with numeric values

    Returns
    -------
    bool

    """

    # 1. make sure keys are the same
    if d1.keys() != d2.keys():
        return False

    # 2. for each key check if values are close
    return all(isclose(d1[key], d2[key],
                       rel_tol=REL_TOL,
                       abs_tol=ABS_TOL) for key in d1.keys())


# represent result as dictionary for comparison with control dict
def result_to_dict(result_dataframe):
    result = result_dataframe.copy()
    result['bi-gram'] = result[event_col] + '~~~' + \
                        result[next_event_col]
    return dict(zip(result['bi-gram'], result['edge_weight']))


def no_norm_by_events(test_dataset):
    """
    norm_type=None,
    weight_col=None
    """

    edgelist = test_dataset.rete.get_edgelist(norm_type=None,
                                              weight_col=None)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    result_test = dict(control_dataset['bi-gram'].value_counts())

    return dict_compare(result_rete, result_test)


def full_norm_by_events(test_dataset):
    """
    norm_type='full',
    weight_col=None
    """
    # obtain expected result using rete lib
    edgelist = test_dataset.rete.get_edgelist(norm_type='full',
                                              weight_col=None)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    result_test = dict(control_dataset['bi-gram'].value_counts(normalize=True))

    check_dictionaries = dict_compare(result_rete, result_test)

    # make sure all weights sum to 1:
    check_control_sum = isclose(edgelist['edge_weight'].sum(), 1,
                                rel_tol=REL_TOL, abs_tol=ABS_TOL)

    return all([check_dictionaries, check_control_sum])


def node_norm_by_events(test_dataset):
    """
    norm_type='node',
    weight_col=None
    """
    # obtain expected result using rete lib
    edgelist = test_dataset.rete.get_edgelist(norm_type='node',
                                              weight_col=None)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    node_transitions_counts = control_dataset.groupby(event_col)['bi-gram'].count()
    grouped_control = control_dataset.groupby('bi-gram').agg({event_col: 'first',
                                                              time_col: 'count'}).reset_index()

    grouped_control['node_norm'] = grouped_control[time_col] \
                                   / node_transitions_counts.loc[grouped_control[event_col]].values

    result_test = dict(zip(grouped_control['bi-gram'], grouped_control['node_norm']))

    # are the control and test dicts same?
    check_dictionaries = dict_compare(result_rete, result_test)

    # make sure sum of normalized weights for transitions
    # from each event is equal to 1 (total prob is 1)
    control_sum = edgelist.groupby(event_col)['edge_weight'].sum()
    check_control_sum = all(isclose(x, 1, rel_tol=REL_TOL, abs_tol=ABS_TOL)
                            for x in control_sum)

    return all([check_dictionaries, check_control_sum])


def no_norm_by_users(test_dataset):
    """
    norm_type=None,
    weight_col=INDEX_COL
    """
    edgelist = test_dataset.rete.get_edgelist(norm_type=None,
                                              weight_col=index_col)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    result_test = dict(control_dataset.groupby(['bi-gram'])[index_col].nunique())

    return dict_compare(result_rete, result_test)


def full_norm_by_users(test_dataset):
    """
    norm_type='full',
    weight_col=INDEX_COL
    """
    edgelist = test_dataset.rete.get_edgelist(norm_type='full',
                                              weight_col=index_col)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    result_test = dict(control_dataset.groupby(['bi-gram'])[index_col].nunique()
                       / control_dataset[index_col].nunique())

    return dict_compare(result_rete, result_test)


def node_norm_by_users(test_dataset):
    """
    norm_type='node',
    weight_col=INDEX_COL
    """
    edgelist = test_dataset.rete.get_edgelist(norm_type='node',
                                              weight_col=index_col)
    result_rete = result_to_dict(edgelist)

    # obtain expected result using control dataset
    control_dataset = test_dataset.rete._get_shift()
    control_dataset['bi-gram'] = control_dataset[event_col] + '~~~' + \
                                 control_dataset[next_event_col]

    node_counts = control_dataset.groupby(event_col)[index_col].nunique().to_dict()
    g_test = control_dataset.groupby('bi-gram').agg({event_col: 'first',
                                                     index_col: lambda x: x.nunique()}).reset_index()
    g_test['node_norm'] = g_test[index_col] / g_test[event_col].map(node_counts)
    result_test = dict(zip(g_test['bi-gram'], g_test['node_norm']))

    return dict_compare(result_rete, result_test)



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


from datetime import datetime
import seaborn as sns
import pandas as pd

from .plot_utils import __save_plot__

_ = pd.DataFrame()

def compare(*,
            num_data,
            group_names):

    data1 = pd.DataFrame(data={'data': num_data[0], 'groups': group_names[0]})
    data2 = pd.DataFrame(data={'data': num_data[1], 'groups': group_names[1]})
    combined_stats = pd.concat([data1, data2])

    compare_plot = sns.displot(data=combined_stats, x="data", hue="groups", multiple="dodge")

    compare_plot.set(xlabel=None)

    # move legend outside the box
    #compare_plot.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    plot_name = 'compare_plot_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = _.rete.retention_config['experiments_folder'] + '/' + plot_name
    return compare_plot, plot_name, None, _.rete.retention_config


# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import networkx as nx
import pandas as pd


def get_adjacency(self, *,
                  weight_col=None,
                  norm_type=None):
    """
    Creates edge graph in the matrix format. Row indeces are event_col values,
     from which the transition occured, and columns are events, to
    which the transition occured. The values are weights of the edges defined
    with weight_col and norm_type parameters.

    Parameters
    ----------
    weight_col: str (optional, default=None)
        Aggregation column for transitions weighting. To calculate weights
        as number of transion events use None. To calculate number
        of unique users passed through given transition 'user_id'.
         For any other aggreagtion, like number of sessions, pass the column name.

    norm_type: {None, 'full', 'node'} (optional, default=None)
        Type of normalization. If None return raw number of transtions
        or other selected aggregation column. 'full' - normalized over
        entire dataset. 'node' weight for edge A --> B normalized over
        user in A

    Returns
    -------
    Dataframe with number of columns and rows equal to unique number of
    event_col values.

    Return type
    -----------
    pd.DataFrame
    """
    agg = self.get_edgelist(weight_col=weight_col,
                            norm_type=norm_type)
    graph = nx.DiGraph()
    graph.add_weighted_edges_from(agg.values)
    return nx.to_pandas_adjacency(graph)



import pandas as pd
from os.path import dirname

module_path = dirname(__file__)


def load_simple_shop():
    return pd.read_csv(module_path+'/data/simple-onlineshop.csv')


def load_simple_ab_test():
    data = pd.read_csv(module_path+'/data/ab_test_demo.csv')
    data['transaction_ID'] = data['transaction_ID'].astype(str)
    data.loc[data['transaction_ID'] == 'nan', 'transaction_ID'] = None
    data.loc[data['transaction_ID'].notna(), 'transaction_ID'] = \
        data.loc[data['transaction_ID'].notna(), 'transaction_ID'].apply(lambda x: x.replace('.0', ''))
    return data



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import json
from datetime import datetime

import networkx as nx
import pandas as pd

from retentioneering.visualization import templates
from .plot_utils import __save_plot__, ___DynamicFigureWrapper__


_ = pd.DataFrame()


def _calc_layout(data,
                 node_params,
                 width=500,
                 height=500,
                 **kwargs):
    G = nx.DiGraph()

    G.add_weighted_edges_from(data.loc[:, ['source', 'target', 'weight']].values)

    pos_new = nx.layout.spring_layout(G, k=kwargs.get('k', .1),
                                      iterations=kwargs.get('iterations', 300),
                                      threshold=kwargs.get('nx_threshold', 1e-4),
                                      seed=0)

    min_x = min([j[0] for i, j in pos_new.items()])
    min_y = min([j[1] for i, j in pos_new.items()])
    max_x = max([j[0] for i, j in pos_new.items()])
    max_y = max([j[1] for i, j in pos_new.items()])
    pos_new = {
        i: [(j[0] - min_x) / (max_x - min_x) * (width - 150) + 75,
            (j[1] - min_y) / (max_y - min_y) * (height - 100) + 50]
        for i, j in pos_new.items()
    }
    return pos_new, dict(G.degree)


def _prepare_nodes(data, pos, node_params, degrees):
    node_set = set(data['source']) | set(data['target'])
    max_degree = max(degrees.values())
    nodes = {}
    for idx, node in enumerate(node_set):
        node_pos = pos.get(node)
        nodes.update({node: {
            "index": idx,
            "name": node,
            "x": node_pos[0],
            "y": node_pos[1],
            "type": (node_params.get(node) or "suit").split('_')[0] + '_node',
            "degree": (abs(degrees.get(node, 0))) / abs(max_degree) * 30 + 4
        }})
    return nodes


def _prepare_edges(data, nodes):
    edges = []
    data['weight_norm'] = data['weight'] / data['weight'].abs().max()
    for idx, row in data.iterrows():
        edges.append({
            "source": nodes.get(row.source),
            "target": nodes.get(row.target),
            "weight": row.weight_norm,
            "weight_text": row.weight,
            "type": row['type']
        })
    return edges, list(nodes.values())


def _filter_edgelist(data,
                     thresh,
                     node_params,
                     targets=None):
    if targets is None:
        x = pd.Series(node_params).astype(str).str.contains('target')
        targets = set(x[x].index)
    f = data.weight.abs() >= thresh
    nodes = set(data[f].source) | set(data[f].target)
    f |= (data.source.isin(targets) & data.target.isin(nodes))
    f |= (data.target.isin(targets) & data.source.isin(nodes))
    return data[f].copy()


def _make_json_data(data,
                    node_params,
                    layout_dump,
                    thresh=0.0,
                    width=500,
                    height=500,
                    **kwargs):
    res = {}
    data.columns = ['source', 'target', 'weight']
    data = _filter_edgelist(data, thresh, node_params)

    data["type"] = data.apply(
        lambda x: node_params.get(x.source) if node_params.get(x.source) == 'source' else node_params.get(
            x.target) or 'suit', 1)

    pos, degrees = _calc_layout(data, node_params, width=width, height=height, **kwargs)

    if kwargs.get('node_weights') is not None:
        degrees = kwargs.get('node_weights')

    if layout_dump is not None:
        nodes = _prepare_given_layout(layout_dump, node_params, degrees)
    else:
        nodes = _prepare_nodes(data, pos, node_params, degrees)

    res['links'], res['nodes'] = _prepare_edges(data, nodes)
    return res


def _prepare_node_params(node_params, data):
    if node_params is None:
        _node_params = {
            'positive_target_event': 'nice_target',
            'negative_target_event': 'bad_target',
            'source_event': 'source',
        }
        node_params = {}
        for key, val in _node_params.items():
            name = data.rete.retention_config.get(key)
            if name is None:
                continue
            node_params.update({name: val})
    return node_params


def _prepare_layout(layout):
    nodes = {}
    for i in layout:
        nodes.update({i['name']: i})
    return nodes


def _prepare_given_layout(nodes_path, node_params, degrees):
    if type(nodes_path) is str:
        with open(nodes_path, encoding='utf-8') as f:
            nodes = json.load(f)
    else:
        nodes = nodes_path
    if type(nodes) is list:
        nodes = _prepare_layout(nodes)
    return nodes


@__save_plot__
def graph(data, *,
          node_params=None,
          thresh=0.0,
          width=800,
          height=500,
          interactive=True,
          layout_dump=None,
          show_percent=True,
          plot_name=None,
          node_weights=None,
          **kwargs):
    """
    Create interactive graph visualization. Each node is a unique ``event_col`` value, edges are transitions between events and edge weights are calculated metrics. By default, it is a percentage of unique users that have passed though a particular edge visualized with the edge thickness. Node sizes are  Graph loop is a transition to the same node, which may happen if users encountered multiple errors or made any action at least twice.
    Graph nodes are movable on canvas which helps to visualize user trajectories but is also a cumbersome process to place all the nodes so it forms a story.
    That is why IFrame object also has a download button. By pressing it, a JSON configuration file with all the node parameters is downloaded. It contains node names, their positions, relative sizes and types. It it used as ``layout_dump`` parameter for layout configuration. Finally, show weights toggle shows and hides edge weights.
    Parameters
    ---------
    data: pd.DataFrame
        Graph in edgelist form.
    node_params: dict, optional
        Event mapping describing which nodes or edges should be highlighted by different colors for better visualisation. Dictionary keys are ``event_col`` values, while keys have the following possible values:
            - ``bad_target``: highlights node and all incoming edges with red color;
            - ``nice_target``: highlights node and all incoming edges with green color;
            - ``bad_node``: highlights node with red color;
            - ``nice_node``: highlights node with green color;
            - ``source``: highlights node and all outgoing edges with yellow color.
        Example ``node_params`` is shown below:
        ```
        {
            'lost': 'bad_target',
            'purchased': 'nice_target',
            'onboarding_welcome_screen': 'source',
            'choose_login_type': 'nice_node',
            'accept_privacy_policy': 'bad_node',
        }
        ```
        If ``node_params=None``, it will be constructed from ``retention_config`` variable, so that:
        ```
        {
            'positive_target_event': 'nice_target',
            'negative_target_event': 'bad_target',
            'source_event': 'source',
        }
        ```
    thresh: float, optional
        Minimal edge weight value to be rendered on a graph. If a node has no edges of the weight >= ``thresh``, then it is not shown on a graph. It is used to filter out rare event and not to clutter visualization. Default: ``0.05``
    width: float, optional
        Width of plot in pixels. Default: ``500``
    height: float, optional
        Height of plot in pixels. Default: ``500``
    interactive: bool, optional
        If ``True``, then plots graph visualization in interactive session (Jupyter notebook). Default: ``True``
    layout_dump: str, optional
        Path to layout configuration file relative to current directory. If defined, uses configuration file as a graph layout. Default: ``None``
    show_percent: bool, optional
        If ``True``, then all edge weights are converted to percents by multiplying by 100 and adding percentage sign. Default: ``True``
    Returns
    -------
    Saves webpage with JS graph visualization to ``retention_config.experiments_folder``.
    Return type
    -------
    HTML
    """
    scale = data['edge_weight'].abs().max()

    if node_params is None:
        node_params = _prepare_node_params(node_params, data)
    res = _make_json_data(data,
                          node_params,
                          layout_dump,
                          thresh=thresh,
                          width=round(width - width / 3),
                          height=round(height - height / 3),
                          node_weights=node_weights,
                          **kwargs)

    res['node_params'] = node_params

    show = 0
    if show_percent:
        show = 1

    dump = 1 if (layout_dump is not None) or (kwargs.get('is_model', False)) else 0

    __TEMPLATE__ = templates.__OLD_TEMPLATE__ if kwargs.get('use_old', False) else templates.__TEMPLATE__

    x = __TEMPLATE__.format(
            width=width,
            height=height,
            links=json.dumps(res.get('links')).encode('latin1').decode('utf-8'),
            node_params=json.dumps(node_params).encode('latin1').decode('utf-8'),
            nodes=json.dumps(res.get('nodes')).encode('latin1').decode('utf-8'),
            show_percent=show,
            layout_dump=dump,
            thresh=thresh/scale,
            scale=scale
    )

    plot_name = 'graph_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.html'
    plot_name = _.rete.retention_config['experiments_folder'] + '/' + plot_name

    return (
        ___DynamicFigureWrapper__(x, interactive, width, height, res),
        plot_name,
        plot_name,
        data.rete.retention_config
    )





# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import json
from datetime import datetime
import matplotlib.pyplot as plt

import seaborn as sns
from functools import wraps
from IPython.display import IFrame, display, HTML


def __save_plot__(func):
    @wraps(func)
    def save_plot_wrapper(*args, **kwargs):
        sns.mpl.pyplot.show()
        sns.mpl.pyplot.close()
        res = func(*args, **kwargs)
        if len(res) == 2:
            (vis_object, name), res, cfg = res, None, None
        elif len(res) == 3:
            (vis_object, name, res), cfg = res, None
        else:
            vis_object, name, res, cfg = res
        idx = 'id: ' + str(int(datetime.now().timestamp()))
        coords = vis_object.axis()

        vis_object.get_figure().savefig(name, bbox_inches="tight", dpi=cfg.get('save_dpi') or 200)
        return res

    return save_plot_wrapper


class ___FigureWrapper__(object):
    def __init__(self, fig):
        self.fig = fig

    def get_figure(self):
        return self.fig

    def axis(self):
        if len(self.fig.axes) > 1:
            x = self.fig.axes[1].axis()
        else:
            x = self.fig.axes[0].axis()
        return (x[0] / 64, x[0] + (x[1] - x[0]) / 50, x[2] / 1.5, x[3] / 1.5)

    def text(self, *args, **kwargs):
        self.fig.text(*args, **kwargs)


class ___DynamicFigureWrapper__(object):
    def __init__(self, fig, interactive, width, height, links):
        self.fig = fig
        self.interactive, self.width, self.height = interactive, width, height
        self.links = links

    def get_figure(self):
        savefig = __SaveFigWrapper__(self.fig, self.interactive, self.width, self.height)
        return savefig

    def text(self, x, y, text, *args, **kwargs):
        # parts = self.fig.split('<main>')
        # res = parts[:1] + [f'<p>{text}</p>'] + parts[1:]
        # self.fig = '\n'.join(res)
        pass

    def get_raw(self, path):
        base = '.'.join(path.split('.')[:-1])
        with open(base + '_config.json', 'w', encoding="utf-8") as f:
            json.dump(self.links, f)
        return base + '_config.json'

    @staticmethod
    def axis():
        return 4 * [0]


class __SaveFigWrapper__(object):
    def __init__(self, data, interactive=True, width=1000, height=700):
        self.data = data
        self.interactive = interactive
        self.width = width
        self.height = height

    def savefig(self, name, **kwargs):
        with open(name, 'w', encoding="utf-8") as f:
            f.write(self.data)
        if self.interactive:
            display(IFrame(name, width=self.width + 200, height=self.height + 200))



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import numpy as np

from retentioneering.visualization import plot_clusters
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture


def get_clusters(self, *,
                 feature_type='tfidf',
                 ngram_range=(1, 1),
                 n_clusters=8,
                 method='kmeans',
                 plot_type=None,
                 refit_cluster=True,
                 targets=None,
                 **kwargs):
    """
    Cluster users in the dataset according to their behavior.

    Parameters
    ----------
    feature_type: str (optional, default 'tfidf')
        Type of vectorizer to user to convert sequences of events to numerical vectors.
        Currently supports: {'tfidf', 'count', 'frequency', 'binary'}.

    ngram_range: tuple (optional, default (1, 1))
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n will
        be used. For example an ngram_range of (1, 1) means only unigrams, (1, 2)
        means unigrams and bigrams, and (2, 2) means only bigrams.

    n_clusters: int (optional, default 8)
        Number of clusters to be identified.

    method: str (optional, default 'kmeans')
        Clustering method to use. Currently supports: 'kmeans' and 'gmm'.

    plot_type: str (optional, default None)
        Type of cluster statistics overview graph to plot after clustering.
        Currently supports: 'cluster_bar'

    targets: list (optional, default None)
        List of target events to be
        Only applies if plot_type = 'cluster_bar'

    refit_cluster: bool (optional, default True)
        If False, then cached results of previous clustering are used.
        (from .cluster_mapping attribute). If True recalculates clustering.

    Returns
    -------
    Array of clusters as .cluster_mapping attribute

    Return type
    -----------
    np.array
    """
    index_col = self.retention_config['user_col']
    event_col = self.retention_config['event_col']

    # obtain vectorized features
    if hasattr(self, 'datatype') and self.datatype == 'features':
        features = self._obj.copy()
    else:
        features = self.extract_features(feature_type=feature_type,
                                         ngram_range=ngram_range)

    # obtain clusters
    if not hasattr(self, 'clusters') or refit_cluster:
        clusterer = globals()['_'+method]
        self.clusters = clusterer(features,
                                  n_clusters=n_clusters,
                                  **kwargs)
        _create_cluster_mapping(self, features.index.values)

    # init and obtain bool vector for targets:
    targets_bool = [np.array([False] * len(self.clusters))]
    target_names = [' ']

    if targets is not None:
        targets_bool = []
        target_names = []

        # format targets to list of lists:
        for n, i in enumerate(targets):
            if type(i) != list:
                targets[n] = [i]
        for t in targets:
            # get name
            target_names.append('CR: ' + ' '.join(t))
            # get bool vector
            targets_bool.append((self._obj
                                 .groupby(index_col)[event_col]
                                 .apply(lambda x: bool(set(t) & set(x)))
                                 .to_frame()
                                 .sort_index()[event_col]
                                 .values))

    if plot_type:
        func = getattr(plot_clusters, plot_type)
        func(
            features,
            clusters=self.clusters,
            target=targets_bool,
            target_names=target_names,
            refit=refit_cluster,
            **kwargs
        )

    return self.clusters


def filter_cluster(self, cluster_name):
    """
    Filters dataset against one or several clusters.

    Parameters
    ----------
    cluster_name: int or list
        Cluster ID or list of cluster IDs for filtering.

    Returns
    -------
    Filtered dataset as pandas dataframe

    Return type
    -----------
    pd.Dataframe
    """
    index_col = self.retention_config['user_col']
    ids = []
    if type(cluster_name) is list:
        for i in cluster_name:
            ids.extend(self.cluster_mapping[i])
    else:
        ids = self.cluster_mapping[cluster_name]
    return self._obj[self._obj[index_col].isin(ids)].copy().reset_index(drop=True)


def cluster_event_dist(self,
                       cl1,
                       cl2=None, *,
                       top_n=8,
                       weight_col=None,
                       targets=[]):
    """
    Plots distribution of top_n events in cluster cl1 compared vs
    entire dataset or in cluster cl2.

    Parameters
    ----------
    cl1: int
        ID of the cluster to compare.
    cl2: int, (optional, default None)
        ID of the second cluster to compare with top events from first
        cluster. If None, then compares with entire dataset.
    top_n: int, (optional, default 8)
        Number of top events.
    weight_col: str (optional, default None)
        If None distribution will be compared based on events occurrences in
        datasets. If weight_col is specified, percentages of users
        (column name specified by parameter weight_col) who have particular
        events will be plotted.
    targets: list of str (optional, default [])
        List of event names always to include for comparison regardless
        of the parameter top_n value. Target events will appear in the same
        order as specified

    Returns
    -------
    Plots distribution barchart
    """
    event_col = self.retention_config['event_col']
    index_col = self.retention_config['user_col']

    clus = self.filter_cluster(cl1)

    if weight_col is not None:
        clus.drop_duplicates(subset=[event_col, weight_col], inplace=True)

        top_cluster = (clus[event_col]
                       .value_counts() / clus[weight_col].nunique())

    else:
        top_cluster = (clus[event_col]
                       .value_counts(normalize=True))

    # add zero events for missing targets
    for evnt in set(targets) - set(top_cluster.index):
        top_cluster.loc[evnt] = 0

    # create events order: top_n non-target events + targets:
    evnts_to_keep = list(filter(lambda x: x not in targets, top_cluster.index))[:top_n]
    target_separator_position = len(evnts_to_keep)
    evnts_to_keep += list(targets)
    top_cluster = top_cluster.loc[evnts_to_keep].reset_index()

    if cl2 is None:
        clus2 = self._obj.copy()
    else:
        clus2 = self.filter_cluster(cl2)

    if weight_col is not None:
        clus2.drop_duplicates(subset=[event_col, weight_col], inplace=True)
        # get events distribution from cluster 2:
        top_all = (clus2[event_col]
                   .value_counts()/clus2[weight_col].nunique())
    else:
        # get events distribution from cluster 2:
        top_all = (clus2[event_col]
                   .value_counts(normalize=True))

    # make sure top_all has all events from cluster 1
    for evnt in set(top_cluster['index']) - set(top_all.index):
        top_all.loc[evnt] = 0

    # keep only top_n events from cluster1
    top_all = top_all.loc[top_cluster['index']].reset_index()

    top_all.columns = [event_col, 'freq', ]
    top_cluster.columns = [event_col, 'freq', ]

    top_all['hue'] = 'all' if cl2 is None else f'cluster {cl2}'
    top_cluster['hue'] = f'cluster {cl1}'

    plot_clusters.cluster_event_dist(
        top_all.append(top_cluster, ignore_index=True, sort=False),
        event_col=event_col,
        cl1=cl1,
        sizes=[
            clus[index_col].nunique() / self._obj[index_col].nunique(),
            clus2[index_col].nunique() / self._obj[index_col].nunique(),
        ],
        cl2=cl2,
        weight_col=weight_col,
        target_pos=target_separator_position,
        targets=targets
    )


def _create_cluster_mapping(self, ids):
    self.cluster_mapping = {}
    for cluster in set(self.clusters):
        self.cluster_mapping[cluster] = ids[self.clusters == cluster].tolist()


def _kmeans(data, *,
           n_clusters=8,
           random_state=0,
           **kwargs):
    """
    Finds cluster of users in data.

    Parameters
    -------
    n_clusters
    data: pd.DataFrame
        Dataframe with features for clustering indexed as in ``retention_config.index_col``
    max_n_clusters: int, optional
        Maximal number of clusters for automatic selection for number of clusters. If ``None``, then uses n_clusters from arguments. Default: `None```
    use_csi: bool, optional
        If ``True``, then cluster stability index will be calculated. IMPORTANT: it may take a lot of time. Default: ``True``
    random_state: int, optional
        Random state for KMeans clusterer. Default: ``0``
    kwargs: optional
        Parameters for ``sklearn.cluster.KMeans``

    Returns
    -------
    Array of clusters

    Return type
    -------
    np.array
    """
    km = KMeans(random_state=random_state,
                n_clusters=n_clusters)

    cl = km.fit_predict(data.values)

    return cl


def _gmm(data, *,
        n_clusters=8,
        random_state=0,
        **kwargs):
    """
    Finds cluster of users in data using Gaussian Mixture Models.

    Parameters
    --------
    data: pd.DataFrame
        Dataframe with features for clustering indexed as in ``retention_config.index_col``
    max_n_clusters: int, optional
        Maximal number of clusters for automatic selection for number of clusters. If ``None``, then uses ```n_clusters`` from arguments. Default: `None```
    use_csi: bool, optional
        If ``True``, then cluster stability index will be calculated. IMPORTANT: it may take a lot of time. Default: ``True``
    random_state: int, optional
        Random state for GaussianMixture clusterer.
    kwargs: optional
        Parameters for ``sklearn.mixture.GaussianMixture``

    Returns
    --------
    Array of clusters

    Return type
    --------
    np.array
    """

    km = GaussianMixture(random_state=random_state,
                         n_components=n_clusters)

    cl = km.fit_predict(data.values)

    return cl




# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from .base_dataset import BaseDataset

from .get_edgelist import get_edgelist
from .get_adjacency import get_adjacency
from .step_matrix import step_matrix
from .get_clusters import get_clusters, filter_cluster, cluster_event_dist
from .plot_graph import plot_graph
from .project import project
from .extract_features import extract_features
from .compare import compare
from .funnel import funnel


BaseDataset.get_edgelist = get_edgelist
BaseDataset.get_adjacency = get_adjacency
BaseDataset.step_matrix = step_matrix
BaseDataset.get_clusters = get_clusters
BaseDataset.project = project
BaseDataset.filter_cluster = filter_cluster
BaseDataset.plot_graph = plot_graph
BaseDataset.extract_features = extract_features
BaseDataset.compare = compare
BaseDataset.cluster_event_dist = cluster_event_dist
BaseDataset.funnel = funnel



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import pandas as pd
import numpy as np
import umap.umap_ as umap
from sklearn.manifold import TSNE
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer


_embedding_types = {'tfidf', 'count', 'binary', 'frequency'}


def extract_features(self, *,
                     feature_type='tfidf',
                     ngram_range=(1, 1)):
    """
    User trajectories vectorizer.

    Parameters
    ----------
    feature_type: str, (optional, default 'tfidf')
        Type of vectorizer. Available vectorization methods:
        {'tfidf', 'count', 'binary', 'frequency'}

    ngram_range: tuple, (optional, default (1,1))
        The lower and upper boundary of the range of n-values for different
        word n-grams or char n-grams to be extracted. For example
        ngram_range=(1, 1) means only single events, (1, 2) means single events
        and bigrams.

    Returns
    -------
    Encoded user trajectories

    Return type
    -----------
    pd.DataFrame of (number of users, number of unique events | event n-grams)
    """
    if feature_type not in _embedding_types:
        raise ValueError("Unknown feature type: {}.\nPlease choose one from {}".format(
            feature_type,
            ' '.join(_embedding_types)
        ))

    tmp = self._obj.copy()

    res = _embedder(tmp,
                    feature_type=feature_type,
                    ngram_range=ngram_range)

    return res


def _embedder(data, *,
              feature_type,
              ngram_range=(1, 1)):
    """
    Similar to ``frequency_embedder()``, but normalizes event frequencies with inversed document frequency.

    Parameters
    --------
    data: pd.DataFrame
        Clickstream dataset.
    ngram_range: tuple, optional
        Range of ngrams to use in feature extraction. Default: ``(1, 1)``
    index_col: str, optional
        Name of custom index column, for more information refer to ``init_config``. For instance, if in config you have defined ``index_col`` as ``user_id``, but want to use function over sessions. By default the column defined in ``init_config`` will be used as ``index_col``.
    event_col: str, optional
        Name of custom event column, for more information refer to ``init_config``. For instance, you may want to aggregate some events or rename and use it as new event column. By default the column defined in ``init_config`` will be used as ``event_col``.

    Returns
    --------
    Dataframe with ``index_col`` vectorized by TF-IDF of events.

    Return type
    -------
    pd.DataFrame
    """
    index_col = data.rete.retention_config['user_col']
    event_col = data.rete.retention_config['event_col']

    corpus = data.groupby(index_col)[event_col].apply(
        lambda x: '~~'.join([el.lower() for el in x])
    )

    if feature_type == 'tfidf':
        vectorizer = TfidfVectorizer(ngram_range=ngram_range, token_pattern = '[^~]+').fit(corpus)
    elif feature_type in {'count', 'frequency'}:
        vectorizer = CountVectorizer(ngram_range=ngram_range, token_pattern='[^~]+').fit(corpus)
    elif feature_type == 'binary':
        vectorizer = CountVectorizer(ngram_range=ngram_range, token_pattern='[^~]+', binary=True).fit(corpus)

    cols = [dict_key[0] for dict_key in sorted(vectorizer.vocabulary_.items(), key=lambda x: x[1])]
    vec_data = pd.DataFrame(index=sorted(data[index_col].unique()),
                            columns=cols,
                            data=vectorizer.transform(corpus).todense())

    # normalize if frequency:
    if feature_type == 'frequency':
        vec_data = vec_data.div(vec_data.sum(axis=1), axis=0).fillna(0)

    setattr(vec_data.rete, 'datatype', 'features')
    return vec_data


def _learn_tsne(data, **kwargs):
    """
    Calculates TSNE transformation for given matrix features.

    Parameters
    --------
    data: np.array
        Array of features.
    kwargs: optional
        Parameters for ``sklearn.manifold.TSNE()``

    Returns
    -------
    Calculated TSNE transform

    Return type
    -------
    np.ndarray
    """

    TSNE_PARAMS = ['angle', 'early_exaggeration', 'init', 'learning_rate', 'method', 'metric',
            'min_grad_norm', 'n_components', 'n_iter', 'n_iter_without_progress', 'n_jobs',
            'perplexity', 'verbose']

    kwargs = {k: v for k, v in kwargs.items() if k in TSNE_PARAMS}
    res = TSNE(random_state=0, **kwargs).fit_transform(data.values)
    return pd.DataFrame(res, index=data.index.values)


def _learn_umap(data, **kwargs):
    """
    Calculates UMAP transformation for given matrix features.

    Parameters
    --------
    data: np.array
        Array of features.
    kwargs: optional
        Parameters for ``umap.UMAP()``

    Returns
    -------
    Calculated UMAP transform

    Return type
    -------
    np.ndarray
    """
    reducer = umap.UMAP()
    _umap_filter = reducer.get_params()
    kwargs = {k: v for k, v in kwargs.items() if k in _umap_filter}
    embedding = umap.UMAP(random_state=0, **kwargs).fit_transform(data.values)
    return pd.DataFrame(embedding, index=data.index.values)




# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from scipy.stats import ks_2samp, mannwhitneyu

from ...visualization import plot_compare
from statsmodels.stats.power import TTestIndPower
from statsmodels.stats.weightstats import ttest_ind
import numpy as np
import math

TESTS_LIST = ['ks_2samp', 'mannwhitneyu']


def compare(self, *,
            groups,
            function,
            test,
            group_names=('group_1', 'group_2'),
            alpha=0.05):
    """
    Tests selected metric between two groups of users.

    Parameters
    ----------
    groups: tuple (optional, default None)
        Must contain tuple of two elements (g_1, g_2): where g_1 and g_2 are collections
        of user_id`s (list, tuple or set).
    function: function(x) -> number
        Selected metrics. Must contain a function wich takes as an argument dataset for
        single user trajectory and returns a single numerical value.
    group_names: tuple (optional, default: ('group_1', 'group_2'))
        Names for selected groups g_1 and g_2.
    test: {‘mannwhitneyu’, 'ttest', ‘ks_2samp’}
        Test the null hypothesis that 2 independent samples are drawn from the same
        distribution. One-sided tests are used, meaning that distributions are compared
        'less' or 'greater'. Rule of thumbs is: for discrete variables (like convertions
        or number of purchase) use Mann-Whitney (‘mannwhitneyu’) test or t-test (‘ttest’).
         For continious variables (like average_check) use Kolmogorov-Smirnov test ('ks_2samp').
    alpha: float (optional, default 0.05)
        Selected level of significance.

    Returns
    -------
    Prints statistical comparison between two groups over selected metric and test

    Plots a distribution for selected metrics for two groups

    """
    # obtain two populations for each group
    index_col = self.retention_config['user_col']
    data = self._obj
    g1 = data[data[index_col].isin(groups[0])].copy()
    g2 = data[data[index_col].isin(groups[1])].copy()

    # obtain two distributions:
    g1_data = g1.groupby(index_col).apply(function).dropna().values
    g2_data = g2.groupby(index_col).apply(function).dropna().values

    # plot graphs
    plot_compare.compare(num_data=(g1_data, g2_data),
                         group_names=group_names)

    # calculate test statistics
    if test:
        print(f"{group_names[0]} (mean \u00B1 SD): {g1_data.mean():.3f} \u00B1 {g1_data.std():.3f}, n = {len(g1_data)}")
        print(f"{group_names[1]} (mean \u00B1 SD): {g2_data.mean():.3f} \u00B1 {g2_data.std():.3f}, n = {len(g2_data)}")

        p_val_norm, power_norm = _stat_test(g1_data, g2_data, test, alpha)
        p_val_rev, power_rev = _stat_test(g2_data, g1_data, test, alpha)

        if p_val_norm < p_val_rev:
            p_val = p_val_norm
            power = power_norm
            label_max = group_names[0]
            label_min = group_names[1]
        else:
            p_val = p_val_rev
            power = power_rev
            label_max = group_names[1]
            label_min = group_names[0]

        print(f"'{label_max}' is greater than '{label_min}' with P-value: {p_val:.5f}")
        print(f"power of the test: {100*power:.2f}%")


def _stat_test(data_max, data_min, test, alpha):
    # calculate effect size
    if (max(data_max) <= 1 and
            min(data_max) >= 0 and
            max(data_min) <= 1 and
            min(data_min) >= 0):
        # if analyze proportions use Cohen's h:
        effect_size = _cohenh(data_max, data_min)
    else:
        # for other variables use Cohen's d:
        effect_size = _cohend(data_max, data_min)

    # calculate power
    power = TTestIndPower().power(effect_size=effect_size,
                                  nobs1=len(data_max),
                                  ratio=len(data_min) / len(data_max),
                                  alpha=alpha,
                                  alternative='larger')

    if test == 'ks_2samp':
        p_val = ks_2samp(data_max, data_min, alternative='less')[1]
    elif test == 'mannwhitneyu':
        p_val = mannwhitneyu(data_max, data_min, alternative='greater')[1]
    elif test == 'ttest':
        p_val = ttest_ind(data_max, data_min, alternative='larger')[1]
    return p_val, power


# function to calculate Cohen's d:
def _cohend(d1, d2):
    n1, n2 = len(d1), len(d2)
    s1, s2 = np.var(d1, ddof=1), np.var(d2, ddof=1)
    s = math.sqrt(((n1 - 1) * s1 + (n2 - 1) * s2) / (n1 + n2 - 2))
    u1, u2 = np.mean(d1), np.mean(d2)
    return (u1 - u2) / s


# function to calculate Cohen's h:
def _cohenh(d1, d2):
    u1, u2 = np.mean(d1), np.mean(d2)
    return 2*(math.asin(math.sqrt(u1)) -
              math.asin(math.sqrt(u2)))



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from retentioneering.visualization import draw_graph
from IPython.display import display, HTML


def plot_graph(self, *,
               targets={},
               weight_col=None,
               norm_type='full',
               layout_dump=None,
               width=800,
               height=500,
               thresh=0):
    """
    Create interactive graph visualization. Each node is a unique event_col
    value, edges are transitions between events and edge weights are calculated
    metrics. By default, it is a percentage of unique users that have passed
    though a particular edge visualized with the edge thickness. Node sizes are
    Graph loop is a transition to the same node, which may happen if users
    encountered multiple errors or made any action at least twice. Graph nodes
    are movable on canvas which helps to visualize user trajectories but is also
    a cumbersome process to place all the nodes so it forms a story.

    That is why IFrame object also has a download button. By pressing it, a JSON
    configuration file with all the node parameters is downloaded. It contains
    node names, their positions, relative sizes and types. It it used as
    layout_dump parameter for layout configuration. Finally, show weights
    toggle shows and hides edge weights.

    Parameters
    ----------
    norm_type: str (optional, default 'full')
        Type of normalization used to calculate weights for graph edges. Possible
        values are:
            * None
            * 'full'
            * 'node'

    weight_col: str (optional, default None)
        Aggregation column for edge weighting. If None, number of events will be
        calculated. For example, can be specified as `client_id` or `session_id`
        if dataframe has such columns.

    targets: dict (optional, default None)
        Event mapping describing which nodes or edges should be highlighted by
        different colors for better visualisation. Dictionary keys are event_col
        values, while keys have the following possible values:
        Example: {'lost': 'red', 'purchased': 'green', 'main': 'source'}

    thresh: float (optional, default 0.01)
        Minimal edge weight value to be rendered on a graph. If a node has no
        edges of the weight >= thresh, then it is not shown on a graph. It
        is used to filter out rare event and not to clutter visualization. Nodes
        specified in targets parameter will be always shown regardless selected
        threshold.

    layout_dump: str (optional, default None)
        Path to layout configuration file relative to current directory. If
        defined, uses configuration file as a graph layout.

    width: int (optional, default 800)
        Width of plot in pixels.

    height: int (optional, default 500)
        Height of plot in pixels.

    Returns
    -------
    Plots IFrame graph of width and height size.
    Saves webpage with JS graph visualization to
    retention_config.experiments_folder.

    Return type
    -----------
    Renders IFrame object and saves graph visualization as HTML in
    experiments_folder of retention_config.
    """

    event_col = self.retention_config['event_col']

    # TODO: change downstream processing
    if targets is not None:
        for k, v in targets.items():
            if v == 'red':
                v = 'bad_target'
            if v == 'green':
                v = 'nice_target'
            targets[k] = v

    node_weights = self._obj[event_col].value_counts().to_dict()
    data = self.get_edgelist(weight_col=weight_col,
                             norm_type=norm_type)

    path = draw_graph.graph(data,
                            node_params=targets,
                            node_weights=node_weights,
                            layout_dump=layout_dump,
                            width=width,
                            height=height,
                            thresh=thresh)

    # if work from google colab user HTML display:
    try:
        import google.colab
        display(HTML(path))
    except:
        pass

    return path



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import os
import pandas as pd
from .core_functions.base_dataset import BaseDataset

config = {'experiments_folder': 'experiments'}

if not os.path.exists(config['experiments_folder']):
    os.mkdir(config['experiments_folder'])


@pd.api.extensions.register_dataframe_accessor("rete")
class RetentioneeringDataset(BaseDataset):

    def __init__(self, pandas_obj):
        super(RetentioneeringDataset, self).__init__(pandas_obj)
        self.retention_config = config


# inform user to update their old notebooks from v1.0.x:
def init_config(*args, **kwargs):
    class VersionError(Exception):
        pass

    raise VersionError('this function was depricated from v2.0.0. ' +
                       'Please refer to documentaion at https://github.com/retentioneering/retentioneering-tools')


# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import itertools
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from .plot_utils import __save_plot__

@__save_plot__
def step_matrix(data, targets=None, *,
                targets_list=None, plot_name=None,
                title='', centered_position=None, precision=2):

    target_cmaps = itertools.cycle(['BrBG', 'PuOr', 'PRGn', 'RdBu'])

    n_rows = 1 + (len(targets_list) if targets_list else 0)
    n_cols = 1

    grid_specs = ({'wspace': 0.08, 'hspace': 0.04,
                  'height_ratios': [data.shape[0], *list(map(len, targets_list))]}
                 if targets is not None else {})

    f, axs = sns.mpl.pyplot.subplots(n_rows, n_cols, sharex=True,

                                     figsize=(round(data.shape[1] * 0.7),
                                              round((len(data) +
                                                     (len(targets) if targets is not None else 0)) * 0.6)),

                                     gridspec_kw=grid_specs)

    heatmap = sns.heatmap(data,
                          yticklabels=data.index,
                          annot=True,
                          fmt=f'.{precision}f',
                          ax=axs[0] if targets is not None else axs,
                          cmap="RdGy",
                          center=0,
                          cbar=False)

    heatmap.set_title(title, fontsize=16)

    if targets is not None:
        for n, i in enumerate(targets_list):
            sns.heatmap(targets.loc[i],
                        yticklabels=targets.loc[i].index,
                        annot=True,
                        fmt=f'.{precision}f',
                        ax=axs[1 + n],
                        cmap=next(target_cmaps),
                        center=0,
                        vmin=min(pd.core.common.flatten(targets.loc[i].values)),
                        vmax=max(pd.core.common.flatten(targets.loc[i].values)) or 1,
                        cbar=False)

        for ax in axs:
            sns.mpl.pyplot.sca(ax)
            sns.mpl.pyplot.yticks(rotation=0)

            # add vertical lines for central step-matrix
            if centered_position is not None:
                ax.vlines([centered_position-0.02, centered_position+0.98],
                          *ax.get_ylim(),
                          colors='Black',
                          linewidth=0.7)

    else:
        sns.mpl.pyplot.sca(axs)
        sns.mpl.pyplot.yticks(rotation=0)
        # add vertical lines for central step-matrix
        if centered_position is not None:
            axs.vlines([centered_position-0.02, centered_position+0.98],
                       *axs.get_ylim(),
                       colors='Black',
                       linewidth=0.7)

    plot_name = plot_name or 'step_matrix_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.rete.retention_config['experiments_folder'] + '/' + plot_name

    return heatmap, plot_name, None, data.rete.retention_config





# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


from math import isclose

import retentioneering
from retentioneering import datasets

# global constants
REL_TOL = 0.0001
ABS_TOL = 0.0001

# import data
retentioneering.config.update({
    'event_col': 'event',
    'event_time_col': 'timestamp',
    'user_col': 'user_id'
})
data = datasets.load_simple_shop()
test_datasets = [dict(test_dataset=data)]


# util function to cycle through parameters for each test
def pytest_generate_tests(metafunc):
    # called once per each test function
    funcarglist = metafunc.cls.params[metafunc.function.__name__]
    argnames = sorted(funcarglist[0])
    metafunc.parametrize(
        argnames, [[funcargs[name] for name in argnames] for funcargs in funcarglist]
    )


class TestStepMatrix:
    # for each test we would like to run all cases
    params = {}.fromkeys(["test_step_matrix",
                          "test_step_matrix_thresh",
                          "test_step_matrix_targets",
                          "test_step_matrix_centered",
                          "test_step_matrix_sorting",
                          "test_step_matrix_diff"], test_datasets)

    def test_step_matrix(self, test_dataset):
        assert step_matrix_test(test_dataset, max_steps=20,
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=1,
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=100,
                                              precision=3,
                                              show_plot=False)

    def test_step_matrix_thresh(self, test_dataset):
        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.5,
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=1,
                                              show_plot=False)

    def test_step_matrix_targets(self, test_dataset):
        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.5,
                                              targets=['payment_done'],
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.5,
                                              targets=['payment_done', '_TEST_EVENT_'],
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              targets=['product1', ['cart', 'payment_done']],
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              targets=['product1', ['cart', 'payment_done']],
                                              accumulated='only',
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              targets=['product1', ['cart', 'payment_done']],
                                              accumulated='both',
                                              show_plot=False)

    def test_step_matrix_centered(self, test_dataset):
        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              centered={'event': 'cart',
                                                        'left_gap': 5,
                                                        'occurrence': 1},
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              centered={'event': 'cart',
                                                        'left_gap': 10,
                                                        'occurrence': 2},
                                              targets=['payment_done'],
                                              show_plot=False)

        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.05,
                                              centered={'event': 'cart',
                                                        'left_gap': 0,
                                                        'occurrence': 1},
                                              targets=['payment_done'],
                                              show_plot=False)

    def test_step_matrix_sorting(self, test_dataset):
        custom_order = ['catalog', 'cart', 'delivery_choice', 'delivery_courier', 'payment_card', 'payment_choice',
                        'payment_done', 'ENDED', 'THRESHOLDED_6']
        assert step_matrix_test(test_dataset, max_steps=20,
                                              thresh=0.5,
                                              centered={'event': 'payment_done',
                                                        'left_gap': 10,
                                                        'occurrence': 1},
                                              targets=[['catalog', 'cart']],
                                              sorting=custom_order,
                                              show_plot=False)

    def test_step_matrix_diff(self, test_dataset):
        assert step_matrix_diff_test(test_dataset, max_steps=20,
                                                   thresh=0.,
                                                   groups=({3744822}, {736567298}),
                                                   show_plot=False)

        g1 = set(test_dataset[test_dataset['event'] == 'payment_done']['user_id'].unique())
        g2 = set(test_dataset['user_id'].unique()) - g1
        assert step_matrix_diff_test(test_dataset, max_steps=20,
                                                   thresh=0.,
                                                   groups=(g1, g2),
                                                   show_plot=False)

        assert step_matrix_diff_test(test_dataset, max_steps=20,
                                                   thresh=0.,
                                                   groups=(g1, g2),
                                                   targets=['payment_done', 'cart'],
                                                   accumulated='both',
                                                   show_plot=False)

        assert step_matrix_diff_test(test_dataset, max_steps=20,
                                                   thresh=0.,
                                                   groups=(g1, g2),
                                                   targets=['payment_done', 'cart'],
                                                   accumulated='both',
                                                   centered={'event': 'delivery_choice',
                                                             'left_gap': 10,
                                                             'occurrence': 1},
                                                   show_plot=False)


def step_matrix_test(test_dataset, **kwargs):
    """
    """
    res_matrix = test_dataset.rete.step_matrix(**kwargs)

    # making sure columns sum up to 1:
    res_list = list(res_matrix.sum())
    return all(isclose(1, i, rel_tol=REL_TOL, abs_tol=ABS_TOL) for i in res_list)


def step_matrix_diff_test(test_dataset, **kwargs):
    """
    """
    res_matrix = test_dataset.rete.step_matrix(**kwargs)

    # making sure columns sum up to 0:
    res_list = list(res_matrix.sum())
    return all(isclose(0, i, rel_tol=REL_TOL, abs_tol=ABS_TOL) for i in res_list)



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md


import plotly.graph_objects as go
from datetime import datetime
import pandas as pd
_ = pd.DataFrame()


def plot_stacked_funnel(groups):
    data = []

    for t in groups.keys():
        trace = go.Funnel(
            name=t,
            y=groups[t]['targets'],
            x=groups[t]['values'],
            textinfo="value+percent initial+percent previous"
        )
        data.append(trace)

    layout = go.Layout(margin={"l": 180, "r": 0, "t": 30, "b": 0, "pad": 0},
                       funnelmode="stack",
                       showlegend=True,
                       hovermode='closest',
                       legend=dict(orientation="v",
                                   bgcolor='#E2E2E2',
                                   xanchor='left',
                                   font=dict(
                                       size=12)
                                   )
                       )

    fig = go.Figure(data, layout)

    plot_name = 'funnel_plot_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.html'
    path = _.rete.retention_config['experiments_folder'] + '/' + plot_name
    fig.write_html(path)

    return fig



# * Copyright (C) 2020 Maxim Godzi, Anatoly Zaytsev, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

import pandas as pd


class BaseDataset(object):
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.retention_config = {}

    def _get_shift(self, *,
                   index_col=None,
                   event_col=None):
        index_col = index_col or self.retention_config['user_col']
        event_col = event_col or self.retention_config['event_col']
        time_col = self.retention_config['event_time_col']

        data = self._obj.copy()
        data.sort_values([index_col, time_col], inplace=True)
        shift = data.groupby(index_col).shift(-1)

        data['next_'+event_col] = shift[event_col]
        data['next_'+str(time_col)] = shift[time_col]

        return data

    def split_sessions(self, *,
                       by_event=None,
                       thresh,
                       eos_event=None,
                       session_col='session_id'):

        session_col_arg = session_col or 'session_id'

        index_col = self.retention_config['user_col']
        event_col = self.retention_config['event_col']
        time_col = self.retention_config['event_time_col']

        res = self._obj.copy()

        if by_event is None:
            res[time_col] = pd.to_datetime(res[time_col])
            if thresh is None:
                # add end_of_session event at the end of each string
                res.sort_values(by=time_col, inplace=True, ascending=False)
                res[hash('session')] = res.groupby(index_col).cumcount()
                res_session_ends = res[(res[hash('session')] == 0)].copy()
                res_session_ends[event_col] = eos_event
                res_session_ends[time_col] = res_session_ends[time_col] + pd.Timedelta(seconds=1)

                res = pd.concat([res, res_session_ends])

                res.sort_values(by=time_col, inplace=True)

            else:
                # split sessions by time thresh:
                # drop end_of_session events if already present:
                if eos_event is not None:
                    res = res[res[event_col] != eos_event].copy()

                res.sort_values(by=time_col, inplace=True)
                shift_res = res.groupby(index_col).shift(-1)

                time_delta = pd.to_datetime(shift_res[time_col]) - pd.to_datetime(res[time_col])
                time_delta = time_delta.dt.total_seconds()

                # get boolean mapper for end_of_session occurrences
                eos_mask = time_delta > thresh

                # add session column:
                res[hash('session')] = eos_mask
                res[hash('session')] = res.groupby(index_col)[hash('session')].cumsum()
                res[hash('session')] = res.groupby(index_col)[hash('session')].shift(1).fillna(0).map(int).map(str)

                # add end_of_session event if specified:
                if eos_event is not None:
                    tmp = res.loc[eos_mask].copy()
                    tmp[event_col] = eos_event
                    tmp[time_col] += pd.Timedelta(seconds=1)

                    res = pd.concat([res, tmp], ignore_index=True)
                    res = res.sort_values(time_col).reset_index(drop=True)

                res[session_col_arg] = res[index_col].map(str) + '_' + res[hash('session')]

        else:
            # split sessions by event:
            res[hash('session')] = res[event_col] == by_event
            res[hash('session')] = res.groupby(index_col)[hash('session')].cumsum().fillna(0).map(int).map(str)
            res[session_col_arg] = res[index_col].map(str) + '_' + res[hash('session')]

        res.drop(columns=[hash('session')], inplace=True)
        if session_col is None and session_col_arg in res.columns:
            res.drop(columns=[session_col_arg], inplace=True)
        return res


